from clease.montecarlo.observers import MCObserver
import numpy as np


class EnergyEvolution(MCObserver):
    """Trace the evolution of energy."""

    def __init__(self, mc, ignore_reset=False):
        self.mc = mc
        self.energies = []
        self.mean_energies = []
        MCObserver.__init__(self)
        self.name = "EnergyEvolution"
        self.ignore_reset = ignore_reset

    def __call__(self, system_changes):
        """Append the current energy to the MC object.

        Parameters:

        system_changes: list
            System changes. See doc-string of
            `clease.montecarlo.observers.MCObserver`
        """
        self.energies.append(self.mc.current_energy + self.mc.energy_bias)
        self.mean_energies.append(self.mc.mean_energy.mean +
                                  self.mc.energy_bias)

    def reset(self):
        """Reset the history."""
        if self.ignore_reset:
            return
        self.energies = []

    def save(self, fname="energy_evolution"):
        """Save the energy evolution in .csv file.

        Parameters:

        fname: str
            File name without the extension (.csv)
        """
        full_fname = fname + '.csv'
        np.savetxt(full_fname, self.energies, delimiter=",")
        print("Energy evolution data saved to {}".format(full_fname))



from itertools import filterfalse, product
from itertools import combinations
import numpy as np
from clease.cluster_fingerprint import ClusterFingerprint
from scipy.spatial import cKDTree as KDTree
from ase.geometry import wrap_positions


class ClusterGenerator(object):
    """
    Class for generating cluster info that is independent of the
    template
    """
    def __init__(self, prim_cell):
        self.prim = prim_cell
        self.shifts = self.prim.get_positions()

    def eucledian_distance_vec(self, x1, x2):
        """
        Eucledian distance between to vectors

        Parameters:
        x1: array int
            First vector
        x2: array int
            Second vector
        """
        cellT = self.prim.get_cell().T
        euc1 = cellT.dot(x1[:3]) + self.shifts[x1[3]]
        euc2 = cellT.dot(x2[:3]) + self.shifts[x2[3]]
        return euc2 - euc1

    def cartesian(self, x):
        cellT = self.prim.get_cell().T
        return cellT.dot(x[:3]) + self.shifts[x[3]]

    def eucledian_distance(self, x1, x2):
        d = self.eucledian_distance_vec(x1, x2)
        return np.sqrt(np.sum(d**2))

    @property
    def shortest_diag(self):
        shortest = None
        cellT = self.prim.get_cell().T
        for w in product([-1, 0, 1], repeat=3):
            if all(x == 0 for x in w):
                continue

            diag = cellT.dot(w)
            length = np.sqrt(np.sum(diag**2))
            if shortest is None or length < shortest:
                shortest = length
        return shortest

    @property
    def num_sub_lattices(self):
        return len(self.prim)

    def sites_within_cutoff(self, cutoff, ref_lattice=0):
        min_diag = self.shortest_diag
        max_int = int(cutoff/min_diag) + 1
        x0 = [0, 0, 0, ref_lattice]
        sites = filterfalse(
            lambda x: (self.eucledian_distance(x0, x) > cutoff
                       or sum(abs(y) for y in x) == 0),
            product(range(-max_int, max_int+1),
                    range(-max_int, max_int+1),
                    range(-max_int, max_int+1),
                    range(self.num_sub_lattices)))
        return sites

    def get_fp(self, X):
        """
        Generate finger print given a position matrix
        """
        X = np.array(X)
        com = np.mean(X, axis=0)
        X -= com
        X = X.dot(X.T)
        diag = np.diagonal(X)
        off_diag = []
        for i in range(1, X.shape[0]):
            off_diag += np.diagonal(X, offset=i).tolist()
        N = X.shape[0]
        assert len(off_diag) == N*(N-1)/2
        inner = np.array(sorted(diag, reverse=True) + sorted(off_diag))
        fp = ClusterFingerprint(list(inner))
        return fp

    def generate(self, size, cutoff, ref_lattice=0):
        clusters = []
        all_fps = []
        sites = self.sites_within_cutoff(cutoff, ref_lattice=ref_lattice)
        v0 = [0, 0, 0, ref_lattice]
        x0 = np.array(self.cartesian(v0))
        for comb in combinations(sites, r=size-1):
            X = [x0] + [self.cartesian(v) for v in comb]
            fp = self.get_fp(X)
            if np.max(np.sqrt(fp.fp[:size])) > cutoff/2.0:
                continue

            # Find the group
            new_item = [v0] + [list(x) for x in comb]
            try:
                group = all_fps.index(fp)
                clusters[group].append(new_item)
            except ValueError:
                # Does not exist, create a new group
                clusters.append([new_item])
                all_fps.append(fp)

        # Order the indices by internal indices
        for cluster in clusters:
            for i, f in enumerate(cluster):
                ordered_f = self._order_by_internal_distances(f)
                cluster[i] = ordered_f
        return clusters, all_fps

    def _get_internal_distances(self, figure):
        """
        Return all the internal distances of a figure
        """
        dists = []
        for x0 in figure:
            d = []
            for x1 in figure:
                dist = self.eucledian_distance(x0, x1)
                dist = dist.round(decimals=6)
                d.append(dist.tolist())
            dists.append(sorted(d, reverse=True))
        return dists

    def _order_by_internal_distances(self, figure):
        """Order the indices by internal distances."""
        dists = self._get_internal_distances(figure)
        zipped = sorted(list(zip(dists, figure)), reverse=True)
        return [x[1] for x in zipped]

    def to_atom_index(self, clusters, template):
        """
        Convert the integer vector representation to an atomic index

        Parameters:
        clusters: list
            List of clusters (in integer vector representation)

        template: Atoms
            Template atoms to use
        """
        cell = template.get_cell()
        pos = template.get_positions()
        tree = KDTree(pos)
        int_clusters = []

        for cluster in clusters:
            int_cluster = []
            for fig in cluster:
                int_fig = []
                for ivec in fig:
                    x = self.cartesian(ivec)
                    x_mat = np.zeros((1, 3))
                    x_mat[0, :] = x
                    x = wrap_positions(x_mat, cell)
                    d, i = tree.query(x)
                    int_fig.append(i[0])
                int_cluster.append(int_fig)
            int_clusters.append(int_cluster)
        return int_clusters

    def equivalent_sites(self, figure):
        """Find the equivalent sites of a figure."""
        dists = self._get_internal_distances(figure)
        equiv_sites = []
        for i in range(len(dists)):
            for j in range(i+1, len(dists)):
                if np.allclose(dists[i], dists[j]):
                    equiv_sites.append((i, j))

        # Merge pairs into groups
        merged = []
        for equiv in equiv_sites:
            found_group = False
            for m in merged:
                if any(x in m for x in equiv):
                    m.update(equiv)
                    found_group = True
            if not found_group:
                merged.append(set(equiv))
        return [list(x) for x in merged]



"""A collection of miscellaneous functions used for Cluster Expansion."""
from itertools import (
    permutations, combinations, product,
    filterfalse, chain
)
import numpy as np
from collections.abc import Iterable
from random import sample
from ase.db import connect
import json
from clease import _logger
from scipy.spatial import cKDTree as KDTree


class ApproxEqualityList(object):
    """
    Wrapper around a list which implements a new comparison operator. If two
    items in the list is equal within a given tolerance, the items are
    considered to be equal.

    Parameters:

    array: list
        List that is wrapped

    tol: float
        Toleracnce for comparison check
    """

    def __init__(self, array, tol=1E-5):
        self.array = array
        self.tol = tol

    def __lt__(self, other):
        for x, y in zip(self.array, other.array):
            if x - y < -self.tol:
                return True
            elif x - y > self.tol:
                return False
        return False


def index_by_position(atoms):
    """Set atomic indices by its position."""
    # add zero to avoid negative zeros
    tags = atoms.get_positions() + 0
    tags = tags.tolist()
    tags = [ApproxEqualityList(x) for x in tags]
    deco = sorted([(tag, i) for i, tag in enumerate(tags)])

    indices = [i for tag, i in deco]
    return indices


def sort_by_position(atoms):
    """Sort atoms by its position."""
    # Return a new Atoms object with sorted atomic order.
    # The default is to order according to chemical symbols,
    # but if *tags* is not None, it will be used instead.
    # A stable sorting algorithm is used.
    indices = index_by_position(atoms)
    return atoms[indices]


def wrap_and_sort_by_position(atoms):
    """Wrap and sort atoms by their positions."""
    atoms.wrap()
    atoms = sort_by_position(atoms)
    return atoms


def shift(array):
    ref = array[-1]
    array[1:] = array[:-1]
    array[0] = ref
    return array


def ndarray2list(data):
    """
    Convert nested lists of a combination of lists and numpy arrays
    to list of lists
    """
    if not isinstance(data, list) and not isinstance(data, np.ndarray):
        return data

    data = list(data)
    for i in range(len(data)):
        data[i] = ndarray2list(data[i])
    return list(data)


def dec_string(deco, equiv_sites):
    """Create the decoration string based on equiv sites."""
    equiv_dec = sorted(equivalent_deco(deco, equiv_sites))
    return ''.join(str(i) for i in equiv_dec[0])


def equivalent_deco(deco, equiv_sites):
    """Generate equivalent decoration numbers based on equivalent sites."""
    if not equiv_sites:
        return [deco]

    perm = []
    for equiv in equiv_sites:
        perm.append(list(permutations(equiv)))

    equiv_deco = []
    for comb in product(*perm):
        order = []
        for item in comb:
            order += list(item)

        orig_order = list(range(len(deco)))
        for i, srt_indx in enumerate(sorted(order)):
            orig_order[srt_indx] = order[i]
        equiv_deco.append([deco[indx] for indx in orig_order])

    unique_deco = []
    for eq_dec in equiv_deco:
        if eq_dec not in unique_deco:
            unique_deco.append(eq_dec)
    return unique_deco


def flatten(x):
    """Flatten list."""
    if isinstance(x, Iterable):
        return [a for i in x for a in flatten(i)]
    else:
        return [x]


def nested_array2list(array):
    """Convert a nested array/tuple to a nested list."""
    if isinstance(array, np.ndarray):
        array = array.tolist()
    else:
        array = list(array)
    try:
        for i in range(len(array)):
            if isinstance(array[i], np.ndarray):
                array[i] = array[i].tolist()
            else:
                array[i] = list(array[i])
    except TypeError:
        pass
    return array


def update_db(uid_initial=None, final_struct=None, db_name=None,
              custom_kvp_init={}, custom_kvp_final={}):
    """Update the database.

    Parameters:

    uid_initial: int
        entry ID of the initial structure in the database

    final_struct: Atoms
        Atoms object with the final structure with a physical
        quantity that needs to be modeled (e.g., DFT energy)

    db_name: str
        Database name

    custom_kvp_init: dict (optional)
        If desired, one can pass additional key-value-pairs for the
        entry containing the initial structure

    custom_kvp_final: dict (optional)
        If desired, one can pass additional key-value-pairs for the
        entry containing the final structure
    """
    from ase.db import connect
    db = connect(db_name)

    init_row = db.get(id=uid_initial)

    # Check if a final structure already exits
    name = init_row.name
    select_cond = [('name', '=', name), ('struct_type', '=', 'final')]
    exist = sum(1 for row in db.select(select_cond))
    if exist >= 1:
        _logger("A structure with 'name'={} and 'struct_type'=final "
                "already exits in DB".format(name))
        return

    # Write the final structure to database
    kvp_final = {'struct_type': 'final', 'name': name}
    kvp_final.update(custom_kvp_final)
    uid_final = db.write(final_struct, key_value_pairs=kvp_final)

    kvp_update_init = {
        'converged': True,
        'started': '',
        'queued': '',
    }
    if kvp_final['struct_type'] == 'final':
        kvp_update_init['final_struct_id'] = uid_final
    kvp_update_init.update(custom_kvp_init)

    # Update info for the initial structure
    db.update(uid_initial, **kvp_update_init)


def exclude_information_entries():
    """Return selection condition to exlcude all entries in the database that
       only contain information about the clusters.
    """
    return [('name', '!=', 'primitive_cell'),
            ('name', '!=', 'template')]


def get_all_internal_distances(atoms, max_dist, ref_indices):
    """Obtain all internal distances of the passed atoms object and return a
       Numpy array containing all the distances sorted in an ascending order.
    """
    tree = KDTree(atoms.get_positions())
    distances = []
    for ind in ref_indices:
        indices = tree.query_ball_point(atoms[ind].position, max_dist)
        dists = atoms.get_distances(ind, indices)
        for d in dists:
            if np.any(np.abs(np.array(distances) - d) < 1E-6):
                continue
            distances.append(d)
    distances = sorted(distances)

    # Make sure that the first element is 0
    assert distances[0] < 1E-6
    return np.array(distances[1:])


def reconfigure(setting, select_cond=None):
    from clease import CorrFunction
    setting.reconfigure_settings()
    CorrFunction(setting).reconfigure_db_entries(select_cond)


def split_dataset(X, y, nsplits=10):
    """Split the dataset such that it can be used for k-fold
        cross validation."""
    from random import shuffle
    indices = list(range(len(y)))
    shuffle(indices)
    partitions = []
    num_validation = int(len(y)/nsplits)

    if num_validation < 1:
        num_validation = 1
    for i in range(nsplits):
        start = i*num_validation
        end = (i+1)*num_validation
        if i == nsplits-1:
            indx = indices[start:]
        else:
            indx = indices[start:end]
        mask = np.zeros(len(y), dtype=np.uint8)
        mask[indx] = 1
        data = {
            "train_X": X[mask == 0, :],
            "train_y": y[mask == 0],
            "validate_X": X[mask == 1, :],
            "validate_y": y[mask == 1]
        }
        partitions.append(data)
    return partitions


def random_validation_set(num=10, select_cond=None, db_name=None):
    """
    Construct a random test set.

    Parameters:

    num: int
        Number of datapoints to include in the test set
    select_cond: list
        Select condition to be used to select items from
        the database. If not given, it will be struct_type='initial',
        converged=True
    db_name str:
        Name of the database
    """

    db = connect(db_name)
    if select_cond is None:
        select_cond = [('struct_type', '=', 'initial'),
                       ('converged', '=', True)]

    if db_name is None:
        raise ValueError("No database provided!")

    all_ids = []
    for row in db.select(select_cond):
        all_ids.append(row.id)
    return sample(all_ids, num)


def exclude_ids(ids):
    """
    Construct a select condition based on the ids passed.

    Parameters:

    ids: list of int
        List of IDs
    """
    return [("id", "!=", x) for x in ids]


def load_settings(fname):
    """Load Cluster Expansion settings from a JSON file.

    Parameters:

    fname: str
        JSON file with the settings information
    """
    with open(fname, 'r') as infile:
        info = json.load(infile)

    if info['classtype'] == 'CEBulk':
        from clease import CEBulk
        return CEBulk.load(fname)
    elif info['classtype'] == 'CECrystal':
        from clease import CECrystal
        return CECrystal.load(fname)

    allowed_class_types = ['CEBulk', 'CECrystal']
    raise ValueError("Could not find matching ClusterExpansionSetting type. "
                     "Ensure that the classtype field in the JSON file is one "
                     "of {}".format(allowed_class_types))


def symbols2integer(basis_functions):
    """Convert each symbol in the basis function
       to a unique integer.

    Parameters:

    basis_functions: list of dict
        The basis function dictionary from ClusterExpansionSetting
    """
    symb_id = {}
    for i, symb in enumerate(basis_functions[0].keys()):
        symb_id[symb] = i
    return symb_id


def bf2npyarray(basis_functions, symb_id):
    """Convert the basis function dictionary to a 2D
        numpy array

    Parameters:

    basis_function: list of dict
        Basis function dictionary from ClusterExpansionSetting

    symb_id: dict
        Dictionary of symbol integer (see symbols2integer)
    """
    bf_npy = np.zeros((len(basis_functions), len(symb_id)))

    for i, bf in enumerate(basis_functions):
        for k, v in bf.items():
            bf_npy[i, symb_id[k]] = v
    return bf_npy


def nested_list2str(nested_list):
    """Convert a nested list to string."""
    return 'x'.join(','.join(str(x) for x in item) for item in nested_list)


def str2nested_list(string):
    """Convert string to nested list."""
    return [list(map(lambda x: int(x), item.split(',')))
            for item in string.split('x')]


def min_distance_from_facet(x, cell):
    """
    Calculate the minimum distance from a point to the cell facet.

    Parameters:

    x: np.array
        Position from which to calculate the minimum distance

    cell: Cell
        Cell of an Atoms object
    """
    dists = []

    for plane in combinations([0, 1, 2], r=2):
        # Unit normal vector
        n = np.cross(cell[plane[0], :], cell[plane[1], :])
        n /= np.sqrt(n.dot(n))

        # Plane with origin in it
        dist = np.abs(n.dot(x))
        dists.append(dist)

        # Opposite facet
        remaining = list(set([0, 1, 2]) - set(plane))[0]
        vec = cell[remaining, :]
        dist = np.abs(n.dot(x - vec))
        dists.append(dist)
    return min(dists)


def trans_matrix_index2tags(tm, tagged_atoms, indices=None):
    """
    Convert from indices to tags

    Parameters:

    tm: list of dict
        Original translation matrix

    tagged_atoms: Atoms
        Atoms with a tag that should be used instead of the
        index

    indices: list of int
        Atom indices corresponding to each row in tm. If None,
        it is assumed that len(tm) == len(tagged_atoms) and each
        row in tm corresponds to the atom with the same index in
        tagged_atoms.
    """
    unique_tags = sorted(list(set(atom.tag for atom in tagged_atoms)))

    if indices is None:
        indices = list(range(len(tm)))

    # Make sure we have a continuous series of tags
    assert len(unique_tags) == max(unique_tags) + 1

    new_tm = [{} for _ in range(len(unique_tags))]
    used_tags = [False for _ in range(len(unique_tags))]

    for i, row in enumerate(tm):
        tag = tagged_atoms[indices[i]].tag
        if used_tags[tag]:
            continue
        new_row = {tagged_atoms[k].tag: tagged_atoms[v].tag
                   for k, v in row.items()}
        used_tags[tag] = True
        new_tm[tag] = new_row
    return new_tm


def indices2tags(supercell, clusters):
    for cluster in clusters:
        for i, figure in enumerate(cluster):
            cluster[i] = [int(supercell[x].tag) for x in figure]
    return clusters


def list2str(array):
    return "-".join(str(x) for x in array)


def factorize(n):
    while n > 1:
        for i in range(2, n + 1):
            if n % i == 0:
                n = int(n/i)
                yield i
                break


def count_atoms(atoms):
    """
    Return dictionary with the number of items of each species
    """
    count = {}
    for s in atoms.symbols:
        count[s] = count.get(s, 0) + 1
    return count


def all_integer_transform_matrices_given_diag(diag):
    rng1 = range(0, diag[0]+1)
    rng2 = rng1
    rng3 = range(0, diag[1]+1)
    for off_diag in product(rng1, rng2, rng3):
        yield np.array(
            [[diag[0], off_diag[0], off_diag[1]],
             [0, diag[1], off_diag[2]],
             [0, 0, diag[2]]]
        )


def all_integer_transform_matrices_per_diag(n):
    """
    Yield all the integer transform matrices
    """
    diags = filterfalse(lambda x: x[0]*x[1]*x[2] != n,
                        product(range(n+1), repeat=3))

    for d in diags:
        yield all_integer_transform_matrices_given_diag(d)


def all_integer_transform_matrices(n):
    return chain(*all_integer_transform_matrices_per_diag(n))


def rotate_cells(cell, target_cell):
    """
    Rotate the cells such that one vector is parallel.
    And the other vector lies in a given plane
    """
    dot_prod_cell = cell.dot(cell.T)
    dot_prod_target_cell = target_cell.dot(target_cell.T)

    # Calculate unit vector cells
    uvec_cell = np.zeros_like(cell)
    uvec_target_cell = np.zeros_like(target_cell)
    for i in range(3):
        uvec_cell[i, :] = cell[i, :]/np.sqrt(dot_prod_cell[i, i])
        uvec_target_cell[i, :] = target_cell[i, :] / \
            np.sqrt(dot_prod_target_cell[i, i])

    # Rotate one vector to be parallel
    v = np.cross(uvec_cell[0, :], uvec_target_cell[0, :])
    c = uvec_cell[0, :].dot(uvec_target_cell[0, :])
    v_cross = np.zeros((3, 3))
    v[0, 1] = -v[2]
    v[0, 2] = v[1]
    v[1, 0] = v[2]
    v[1, 2] = -v[0]
    v[2, 0] = -v[1]
    v[2, 1] = v[0]
    R = np.eye(3) + v_cross + v_cross.dot(v_cross)/(1 + c)
    target_cell = R.dot(target_cell)
    return target_cell


def species_chempot2eci(bf_list, species_chempot):
    """
    Convert chemical potentials given for species to their corresponding
    singlet ECI values.

    Parameters:

    bf_list: list
        List of basis function values for each species

    species_chempot: dict
        Dictionary containing the chemical potential for each species. Note
        that the chemical potential for one species can be set to 0.0 without
        any loss of generality. Hence, the species_chempot should contain
        chemical potentials for all elements except one.
    """
    if len(species_chempot) != len(bf_list):
        msg = 'Inconsistent number of chemical potentials. Basis functions\n'
        msg += '{}. Passed chemical potentials {}'.format(bf_list,
                                                          species_chempot)
        raise ValueError(msg)

    n = len(species_chempot)
    mat = np.zeros((n, n))
    rhs = np.zeros(n)
    row = 0
    for sp, mu in species_chempot.items():
        for col, bf in enumerate(bf_list):
            mat[row, col] = bf[sp]
        rhs[row] = mu
        row += 1
    try:
        eci_chem_pot = np.linalg.solve(mat, rhs)
    except np.linalg.LinAlgError:
        inv_mat = np.linalg.pinv(mat)
        eci_chem_pot = inv_mat.dot(rhs)
    eci_chem_pot = eci_chem_pot.tolist()
    eci_dct = {'c1_{}'.format(i): v for i, v in enumerate(eci_chem_pot)}
    return eci_dct



from clease.montecarlo.observers import MCObserver
import numpy as np


class SiteOrderParameter(MCObserver):
    """
    Detect phase transitions by monitoring the average number of sites that
    are occupied by a different element from the initial structure.

    Parameters:

    atoms: Atoms object
        Atoms object use for Monte Carlo
    """

    def __init__(self, atoms):
        self.atoms = atoms
        self.orig_nums = self.atoms.get_atomic_numbers()
        self.avg_num_changed = 0
        self.avg_num_changed_sq = 0
        self.num_calls = 0
        self.current_num_changed = 0
        self.site_changed = np.zeros(len(self.atoms), dtype=np.uint8)

    def _check_all_sites(self):
        """Check if symbols have changed on all sites."""
        nums = self.atoms.get_atomic_numbers()
        self.current_num_changed = np.count_nonzero(nums != self.orig_nums)
        self.site_changed = (nums != self.orig_nums)

    def reset(self):
        """Resets the tracked data. (Not the original symbols array)."""
        self.avg_num_changed = 0
        self.avg_num_changed_sq = 0
        self.num_calls = 0
        self.current_num_changed = 0
        self._check_all_sites()

    def __call__(self, system_changes):
        """Get a new value for the order parameter.

        system_changes: list
            System changes. See doc-string of
            `clease.montecarlo.observers.MCObserver`
        """

        self.num_calls += 1
        assert self.current_num_changed < len(self.atoms)

        # The point this function is called the atoms object is already
        # updated
        for change in system_changes:
            indx = change[0]
            if self.site_changed[indx]:
                if self.atoms[indx].number == self.orig_nums[indx]:
                    self.current_num_changed -= 1
                    self.site_changed[indx] = False
            else:
                if self.atoms[indx].number != self.orig_nums[indx]:
                    self.current_num_changed += 1
                    self.site_changed[indx] = True
        self.avg_num_changed += self.current_num_changed
        self.avg_num_changed_sq += self.current_num_changed**2

    def get_averages(self):
        """
        Get the average and standard deviation of the number of sites that
        are different from the initial state.
        """
        average = float(self.avg_num_changed)/self.num_calls
        average_sq = float(self.avg_num_changed_sq)/self.num_calls
        var = average_sq - average**2

        # If variance is close to zero it can in some cases by
        # slightly negative. Add a safety check for this
        if var < 0.0:
            var = 0.0
        return {"site_order_average": average, "site_order_std": np.sqrt(var)}



class BasisSpecifiedInManyGroupsError(ValueError):
    pass


def parse_max_cluster_dia(string):
    """Parse max cluster dia."""
    string = string.strip('[()]')
    try:
        diameter = float(string)
        return diameter
    except Exception:
        pass

    try:
        diameter = [float(x) for x in string.split(',')]
        return diameter
    except Exception:
        pass

    msg = "Max cluster dia has to be a float or a comma separated list where "
    msg += "each number \nis a cutoff of for 2-body clusters and beyond "
    msg += "(i.e., 3.0, 4.0, 2.0 when the max cluster size is 4)"
    raise ValueError(msg)


def parse_grouped_basis_elements(string):
    """Parse grouped basis argument."""
    string = string.replace('[', '(')
    string = string.replace(']', ')')
    string = string.replace('(', '')
    split = string.split('),')
    split = [x.replace(')', '') for x in split]
    exc_type = ValueError
    msg = 'Grouped basis has to be a list (0, 1, 2, 3) or (0, 1), (2, 3)'
    try:
        grouped = []
        for item in split:
            grouped.append([int(x) for x in item.split(',')])

        # Check that we have only unique sites
        index_set = set()
        index_list = list()
        for grp in grouped:
            index_set = index_set.union(grp)
            index_list += grp

        if len(index_set) != len(index_list):
            exc_type = BasisSpecifiedInManyGroupsError
            msg = 'Basis specified in multiple groups!'
            raise exc_type(msg)
        return grouped

    except Exception as exc:
        raise exc_type(msg)


def parse_size(string):
    """Parse size argument."""
    # List of 3 numbers
    try:
        size = [int(x) for x in string.split(',')]
        if len(size) != 3:
            raise ValueError()
        return size
    except Exception:
        pass

    # Try 3x3 matrix
    string = string.replace('[', '(')
    string = string.replace(']', ')')
    filtered = string.replace('(', '')
    splitted = filtered.split('),')
    splitted = [x.replace(')', '') for x in splitted]
    try:
        matrix = []
        for x in splitted:
            row = [int(y) for y in x.split(',')]
            if len(row) != 3:
                raise ValueError()
            matrix.append(row)

        if len(matrix) == 1:
            matrix = matrix[0]

        if len(matrix) != 3:
            raise ValueError()

        return matrix

    except Exception as exc:
        pass

    msg = 'Size has to be a list with 3 ints (i.e. 3, 4, 5) or rows of a '
    msg += '3x3 matrix \n(1, 0, -1), (0, 1, 1), (2, 0, 3)'
    raise ValueError(msg)


def parse_elements(string):
    """Parse elements given by user."""
    from ase.data import chemical_symbols
    string = string.replace('[', '(')
    string = string.replace(']', ')')
    try:
        if string.count(')') == 0:
            # Single basis
            symb_list = string.split(',')
            symb_list = [x.strip() for x in symb_list]

            if any(x not in chemical_symbols for x in symb_list):
                raise ValueError()
            return [symb_list]

        filtered = string.replace('(', '')
        filtered = filtered.split('),')
        symb_list = [[x.replace(')', '').strip() for x in sub.split(',')]
                     for sub in filtered]

        for sub in symb_list:
            if any(x not in chemical_symbols for x in sub):
                raise ValueError()
        return symb_list
    except Exception:
        pass

    msg = 'Elements must be a comma-separated chemical symbols '
    msg += '(i.e. Al, Mg, Cu) for single sublattice\nand a list '
    msg += 'of symbols for each sublattice otherwise (i.e. (Al, Mg)'
    msg += ', (Zn, Cu), (X, Zn, Cu))'
    raise ValueError(msg)


def parse_cellpar(string):
    """Parse the cell parameter string."""
    string = string.strip('[()]')
    try:
        cellpar = [float(x) for x in string.split(',')]
        if len(cellpar) != 6:
            raise ValueError()
        return cellpar

    except Exception as exc:
        pass

    msg = 'Cell parameters has to be a list of six numbers (a, b, c, alpha, '
    msg += 'beta, gamma). See ASE doc.'
    raise ValueError(msg)


def parse_cell(string):
    """Parse the cell string."""
    string = string.replace('[', '(')
    string = string.replace(']', ')')

    filtered = string.replace('(', '')
    filtered = filtered.split('),')
    filtered = [[x.replace(')', '').strip() for x in sub.split(',')]
                for sub in filtered]

    cell = []
    try:
        for sub in filtered:
            row = [float(x) for x in sub]
            if len(row) != 3:
                raise ValueError()
            cell.append(row)

        if len(cell) != 3:
            raise ValueError()
        return cell
    except Exception:
        pass

    msg = 'Cell has to be row of a 3x3 matrix (i.e (0.0, 4.0, 4.0), (2.0, 1.0,'
    msg += ' 6.0), (3.0, 5.0, 6.0)'
    raise ValueError(msg)


def parse_coordinate_basis(string):
    string = string.replace('[', '(')
    string = string.replace(']', ')')

    filtered = string.replace('(', '')
    filtered = filtered.split('),')
    filtered = [[x.replace(')', '').strip() for x in sub.split(',')]
                for sub in filtered]

    cell = []
    try:
        for sub in filtered:
            row = [float(x) for x in sub]
            if len(row) != 3:
                raise ValueError()
            cell.append(row)

        return cell
    except Exception:
        pass

    msg = 'Coordinate basis has to be list of coordinates (i.e (1.0, 4.0, 4.0)'
    msg += ', (2.0, 1.0, 6.0)'
    raise ValueError(msg)


def parse_temperature_list(string):
    try:
        temps = [float(x) for x in string.split(',')]
        return temps
    except Exception:
        pass
    msg = 'The temperatures has to be a comma separted list of values'
    raise ValueError(msg)


def parse_comma_sep_list_of_int(string):
    try:
        values = [int(x) for x in string.split(',')]
        return values
    except Exception:
        pass
    msg = 'Could not parse list'
    raise ValueError(msg)


def parse_concentration_list(string):
    try:
        per_basis = [x.replace('(', '').replace(')', '')
                     for x in string.split('),')]
        conc_per_basis = []
        for b in per_basis:
            conc_per_basis.append([float(x) for x in b.split(',')])
        return conc_per_basis
    except:
        pass

    msg = 'Concentration list has to be given as one comma separated list per basis'
    raise ValueError(msg)


def parse_select_cond(txt):
    """
    Parse a select condition given as a string.

    Parameter

    txt: str
        Select condition on the form gen=3,struct_type=initial
    """
    select_cond = []
    known_op = ['>=', '<=', '<', '>', '=']
    for cond in txt.split(','):
        for op in known_op:
            if op in cond:
                split = cond.split(op)
                value = split[1]
                try:
                    float_val = float(value)
                    value = float_val
                except Exception:
                    pass

                select_cond.append((split[0], op, value))
                break
    return select_cond



#!/usr/bin/env python
import sys
import os


def print_help_msg():
    msg = "==================================================\n"
    msg += "'clease' command requires additional argument(s).\n"
    msg += "Currently, allowed arguments are:\n"
    msg += "gui : launch GUI for CLEASE\n"
    msg += "gui-setup : Install additional dependencies needed\n"
    msg += "            for running GUI.\n"
    msg += "==================================================\n"
    print(msg)


# For Windows dependencies see
# https://stackoverflow.com/questions/40769386/kivy-windows-unable-to-find-any-valuable-window-provider-at-all
def install_gui_dependencies(args):
    import subprocess
    subprocess.check_call(['pip', 'install', 'kivy'] + args)

    if os.name == 'nt':
        # Windows dependencies
        pckgs = 'docutils pygments pypiwin32 kivy.deps.sdl2 kivy.deps.glew'
        subprocess.check_call(['pip', 'install'] + pckgs.split() + args)

    url = 'https://github.com/kivy-garden/graph/archive/master.zip'
    subprocess.check_call(['pip', 'install', url] + args)


def main():
    argv = sys.argv
    num_args = len(argv)
    if num_args == 1:
        print_help_msg()
        return

    if argv[1] == 'gui':
        from clease.gui.cleaseGUI import CleaseGUI
        CleaseGUI().run()

    elif argv[1] == 'gui-setup':
        install_gui_dependencies(argv[2:])

    elif argv[1] in ['-help', '-h', '--help', 'help']:
        print_help_msg()
        return

if __name__ == "__main__":
    main()



from clease.montecarlo.observers.mc_observer import MCObserver
from clease.montecarlo.observers.sgc_observer import SGCObserver
from clease.montecarlo.observers.corr_func_observer import CorrelationFunctionObserver
from clease.montecarlo.observers.snapshot_observer import Snapshot
from clease.montecarlo.observers.energy_evolution_obs import EnergyEvolution
from clease.montecarlo.observers.site_order_parameter import SiteOrderParameter
from clease.montecarlo.observers.lowest_energy_observer import LowestEnergyStructure
from clease.montecarlo.observers.diffraction_observer import DiffractionObserver
from clease.montecarlo.observers.energy_plot_updater import EnergyPlotUpdater
from clease.montecarlo.observers.concentration_observer import ConcentrationObserver



"""Definition of ClusterExpansionSetting Class.

This module defines the base-class for storing the settings for performing
Cluster Expansion in different conditions.
"""
import os
from itertools import product
from copy import deepcopy
import numpy as np
from ase.db import connect

from clease import _logger, LogVerbosity, ClusterExtractor
from clease.tools import (wrap_and_sort_by_position, indices2tags,
                          get_all_internal_distances, nested_list2str,
                          trans_matrix_index2tags)
from clease.basis_function import BasisFunction
from clease.template_atoms import TemplateAtoms
from clease.concentration import Concentration
from clease.trans_matrix_constructor import TransMatrixConstructor
from clease import AtomsManager
from clease.name_clusters import name_clusters
from clease.cluster_fingerprint import ClusterFingerprint
from clease.cluster import Cluster
from clease.cluster_list import ClusterList
from clease.template_filters import ValidConcentrationFilter


class ClusterExpansionSetting(object):
    """Base class for all Cluster Expansion settings."""

    def __init__(self, concentration, size=None, supercell_factor=27,
                 db_name='clease.db', max_cluster_size=4,
                 max_cluster_dia=[5.0, 5.0, 5.0], basis_function='polynomial',
                 skew_threshold=4, ignore_background_atoms=False):
        self.kwargs = {'size': size,
                       'supercell_factor': supercell_factor,
                       'db_name': db_name,
                       'max_cluster_size': max_cluster_size,
                       'max_cluster_dia': deepcopy(max_cluster_dia),
                       'ignore_background_atoms': ignore_background_atoms}

        if isinstance(concentration, Concentration):
            self.concentration = concentration
        elif isinstance(concentration, dict):
            self.concentration = Concentration.from_dict(concentration)
        else:
            raise TypeError("concentration has to be either dict or "
                            "instance of Concentration")

        self.kwargs["concentration"] = self.concentration.to_dict()
        self.cluster_list = ClusterList()
        self.basis_elements = deepcopy(self.concentration.basis_elements)
        self.num_basis = len(self.basis_elements)
        self.db_name = db_name
        self.size = to_3x3_matrix(size)

        self.prim_cell = self._get_prim_cell()
        self._tag_prim_cell()
        self._store_prim_cell()

        self.template_atoms = TemplateAtoms(supercell_factor=supercell_factor,
                                            size=self.size,
                                            skew_threshold=skew_threshold,
                                            db_name=self.db_name)
        self.atoms_mng = AtomsManager(None)

        self.max_cluster_size = max_cluster_size
        self.max_cluster_dia = self._format_max_cluster_dia(max_cluster_dia)
        self.all_elements = sorted([item for row in self.basis_elements for
                                    item in row])
        self.ignore_background_atoms = ignore_background_atoms
        self.background_indices = None
        self.num_elements = len(self.all_elements)
        self.unique_elements = sorted(list(set(deepcopy(self.all_elements))))
        self.num_unique_elements = len(self.unique_elements)
        self.index_by_basis = None

        self.index_by_trans_symm = []
        self.ref_index_trans_symm = []
        self.template_atoms_uid = 0

        # Set an active template (this ensures that index_by_basis etc is set)
        self._set_active_template_by_uid(0)

        # Run through the templates and filter out the ones that does not have
        # a valid compostion
        self.template_atoms.apply_filter(ValidConcentrationFilter(self))

        if self.template_atoms.num_templates == 0:
            raise RuntimeError('There are no templates the satisfies the '
                               'constraints')

        for uid in range(self.template_atoms.num_templates):
            self._set_active_template_by_uid(uid)
        # Set the initial template atoms to 0, which is the smallest cell
        self._set_active_template_by_uid(0)
        self._check_cluster_list_consistency()

        unique_element_no_bkg = self.unique_element_without_background()
        if isinstance(basis_function, BasisFunction):
            if basis_function.unique_elements != unique_element_no_bkg:
                raise ValueError("Unique elements in BasiFunction instance "
                                 "is different from the one in settings")
            self.bf_scheme = basis_function

        elif isinstance(basis_function, str):
            if basis_function.lower() == 'polynomial':
                from clease.basis_function import Polynomial
                self.bf_scheme = Polynomial(unique_element_no_bkg)
            elif basis_function.lower() == 'trigonometric':
                from clease.basis_function import Trigonometric
                self.bf_scheme = Trigonometric(unique_element_no_bkg)
            elif basis_function.lower() == "binary_linear":
                from clease.basis_function import BinaryLinear
                self.bf_scheme = BinaryLinear(unique_element_no_bkg)
            else:
                msg = "basis function scheme {} ".format(basis_function)
                msg += "is not supported."
                raise ValueError(msg)
        else:
            raise ValueError("basis_function has to be instance of "
                             "BasisFunction or a string")

        self.spin_dict = self.bf_scheme.spin_dict
        self.basis_functions = self.bf_scheme.basis_functions

        if len(self.basis_elements) != self.num_basis:
            raise ValueError("list of elements is needed for each basis")

        if not os.path.exists(db_name):
            self.create_cluster_list_and_trans_matrix()
            self._store_data()
        else:
            self._read_data()

    @property
    def atoms(self):
        return self.atoms_mng.atoms

    def unique_element_without_background(self):
        """Remove backgound elements."""
        if not self.ignore_background_atoms:
            bg_sym = set()
        else:
            bg_sym = set(x[0] for x in self.basis_elements if len(x) == 1)

            # Remove bg_syms that are also present in basis with more than one
            # element
            for elems in self.basis_elements:
                if len(elems) == 1:
                    continue
                to_be_removed = set()
                for s in bg_sym:
                    if s in elems:
                        to_be_removed.add(s)

                bg_sym -= to_be_removed


        unique_elem = set()
        for x in self.basis_elements:
            unique_elem.update(x)
        return list(unique_elem - bg_sym)

    def _size2string(self):
        """Convert the current size into a string."""
        return nested_list2str(self.size)

    def prepare_new_active_template(self, uid):
        """Prepare necessary data structures when setting new template."""
        self.template_atoms_uid = uid
        atoms, self.size = \
            self.template_atoms.get_atoms(uid, return_size=True)
        self.atoms_mng.atoms = wrap_and_sort_by_position(atoms)

        self.index_by_basis = self._group_index_by_basis()

        self.background_indices = self._get_background_indices()
        self.index_by_trans_symm = \
            self.atoms_mng.group_indices_by_trans_symmetry(self.prim_cell)
        if self.ignore_background_atoms:
            self.index_by_trans_symm = [x for x in self.index_by_trans_symm
                                        if x[0] not in self.background_indices]
        self.num_trans_symm = len(self.index_by_trans_symm)
        self.ref_index_trans_symm = [i[0] for i in self.index_by_trans_symm]

    def _set_active_template_by_uid(self, uid):
        """Set a fixed template atoms object as the active."""
        self.prepare_new_active_template(uid)

        # Read information from database
        # Note that if the data is not found, it will generate
        # the nessecary data structures and store them in the database
        self._read_data()

    def set_active_template(self, size=None, atoms=None,
                            generate_template=False):
        """Set a new template atoms object."""
        if size is not None and atoms is not None:
            raise ValueError("Specify either size or pass Atoms object.")
        if size is not None:
            uid = self.template_atoms.get_uid_with_given_size(
                size=size, generate_template=generate_template)
        elif atoms is not None:
            uid = self.template_atoms.get_uid_matching_atoms(
                atoms=atoms, generate_template=generate_template)
        else:
            uid = self.template_atoms.weighted_random_template()
        self._set_active_template_by_uid(uid)

    def _tag_prim_cell(self):
        """Add a tag to all the atoms in the unit cell to track the index."""
        for atom in self.prim_cell:
            atom.tag = atom.index

    def _store_prim_cell(self):
        """Store unit cell to the database."""
        db = connect(self.db_name)
        shape = self.prim_cell.get_cell_lengths_and_angles()
        for row in db.select(name='primitive_cell'):
            uc_shape = row.toatoms().get_cell_lengths_and_angles()
            if np.allclose(shape, uc_shape):
                return row.id

        uid = db.write(self.prim_cell, name='primitive_cell')
        return uid

    def _get_prim_cell(self):
        raise NotImplementedError("This function has to be implemented in "
                                  "in derived classes.")

    def _format_max_cluster_dia(self, max_cluster_dia):
        """Get max_cluster_dia in numpy array form"""
        # max_cluster_dia is list or array
        if isinstance(max_cluster_dia, (list, np.ndarray)):
            # Length should be either max_cluster_size+1 or max_cluster_size-1
            if len(max_cluster_dia) == self.max_cluster_size + 1:
                for i in range(2):
                    max_cluster_dia[i] = 0.
                max_cluster_dia = np.array(max_cluster_dia, dtype=float)
            elif len(max_cluster_dia) == self.max_cluster_size - 1:
                max_cluster_dia = np.array(max_cluster_dia, dtype=float)
                max_cluster_dia = np.insert(max_cluster_dia, 0, [0., 0.])
            else:
                raise ValueError("Invalid length for max_cluster_dia.")
        # max_cluster_dia is int or float
        elif isinstance(max_cluster_dia, (int, float)):
            max_cluster_dia *= np.ones(self.max_cluster_size - 1, dtype=float)
            max_cluster_dia = np.insert(max_cluster_dia, 0, [0., 0.])
        # Case for *None* or something else
        else:
            raise ValueError("max_cluster_dia must be float, int or list.")
        return max_cluster_dia.round(decimals=3)

    def _get_max_cluster_dia(self, cell, ret_weights=False):
        lengths = []
        weights = []
        for w in product([-1, 0, 1], repeat=3):
            vec = cell.dot(w)
            if w == (0, 0, 0):
                continue
            lengths.append(np.sqrt(vec.dot(vec)))
            weights.append(w)

        # Introduce tolerance to make max distance strictly smaller than half
        # of the shortest cell dimension
        tol = 2 * 10**(-3)
        min_length = min(lengths) / 2.0
        min_length = min_length.round(decimals=3) - tol

        if ret_weights:
            min_indx = np.argmin(lengths)
            return min_length, weights[min_indx]
        return min_length

    def _get_background_indices(self):
        """Get indices of the background atoms."""
        # check if any basis consists of only one element type
        basis = [i for i, b in enumerate(self.basis_elements) if len(b) == 1]

        bkg_indices = []
        for b_indx in basis:
            bkg_indices += self.index_by_basis[b_indx]
        return bkg_indices

    def _get_atoms(self):
        """Create atoms with a user-specified size."""
        atoms = self.prim_cell.copy() * self.size
        return wrap_and_sort_by_position(atoms)

    def _check_max_cluster_dia(self, internal_distances):
        """
        Check that the maximum cluster diameter does not exactly correspond
        to an internal distance as this can lead to round off errors
        """
        for dia in self.max_cluster_dia:
            if np.min(np.abs(internal_distances - dia) < 1E-6):
                raise ValueError(
                    "One of the maximum cluster diameters correspond"
                    "to an internal distance. Try to increase the max"
                    "cluster diameter a tiny bit (for instance 4.0 -> 4.01"
                )

    def _get_supercell(self):
        # for atom in atoms:
        #     atom.tag = atom.index
        supercell = self.atoms_mng.close_to_cubic_supercell()
        max_cluster_dia_in_sc = \
            self._get_max_cluster_dia(supercell.get_cell().T)

        # Make large enough supercell (4 times max_cluster_dia can go inside)
        scale = int(4*np.max(self.max_cluster_dia)/max_cluster_dia_in_sc)
        if scale < 1:
            scale = 1
        supercell = supercell*(scale, scale, scale)
        supercell = wrap_and_sort_by_position(supercell)
        ref_indices = \
            self.atoms_mng.corresponding_indices(self.ref_index_trans_symm,
                                                 supercell)

        # Calculate the center of atomic positions in a supercell
        pos = supercell.get_positions()
        com = np.mean(pos, axis=0)

        # Calculate the center all the reference indices
        com_ref = np.mean(pos[ref_indices, :], axis=0)

        # Translate center of reference indices
        supercell.translate(com - com_ref)
        supercell.wrap()
        return supercell, ref_indices

    def _create_cluster_list(self):
        """Generate information for clusters."""
        self.cluster_list.clear()
        supercell, ref_indices = self._get_supercell()
        supercell.info['distances'] = \
            get_all_internal_distances(supercell, max(self.max_cluster_dia),
                                       ref_indices)
        self._check_max_cluster_dia(supercell.info['distances'])

        bkg_sc_indices = []
        if self.ignore_background_atoms:
            sc_manager = AtomsManager(supercell)
            bkg_sc_indices = \
                sc_manager.single_element_sites(self.basis_elements)

        extractor = ClusterExtractor(supercell)
        for size in range(2, self.max_cluster_size+1):
            all_clusters = []
            all_equiv_sites = []
            fingerprints = []
            for ref_indx in ref_indices:
                clusters = extractor.extract(ref_indx=ref_indx, size=size,
                                             cutoff=self.max_cluster_dia[size],
                                             ignored_indices=bkg_sc_indices)
                equiv_sites = \
                    [extractor.equivalent_sites(c[0]) for c in clusters]
                all_equiv_sites.append(equiv_sites)
                clusters = indices2tags(supercell, clusters)
                fingerprints += extractor.inner_prod
                all_clusters.append(clusters)
            names = name_clusters(fingerprints)
            self._update_cluster_list(names, all_clusters, all_equiv_sites,
                                      fingerprints, size)

        # Update with empty info
        for cluster in self.empty_clusters:
            self.cluster_list.append(cluster)

        # Update with singlet info
        for cluster in self.point_clusters:
            self.cluster_list.append(cluster)

    def _update_cluster_list(self, names, clusters, equiv_sites, fingerprints,
                             size):
        counter = 0
        for trans_symm in range(len(clusters)):
            for cluster, equiv in zip(clusters[trans_symm],
                                      equiv_sites[trans_symm]):
                clst = Cluster(names[counter], size,
                               2*np.sqrt(fingerprints[counter][0]),
                               fingerprints[counter],
                               self.ref_index_trans_symm[trans_symm],
                               cluster, equiv, trans_symm)
                self.cluster_list.append(clst)
                counter += 1

    @property
    def empty_clusters(self):
        empty = []
        num_trans_symm = len(self.ref_index_trans_symm)
        for symm in range(num_trans_symm):
            empty.append(
                Cluster('c0', 0, 0.0, ClusterFingerprint([0.0]),
                        self.ref_index_trans_symm[symm], [], [], symm)
            )
        return empty

    @property
    def point_clusters(self):
        point = []
        num_trans_symm = len(self.ref_index_trans_symm)
        for symm in range(num_trans_symm):
            point.append(
                Cluster('c1', 1, 0.0, ClusterFingerprint([1.0]),
                        self.ref_index_trans_symm[symm], [], [], symm)
            )
        return point

    @property
    def multiplicity_factor(self):
        """Return the multiplicity factor of each cluster."""
        num_sites_in_group = [len(x) for x in self.index_by_trans_symm]
        return self.cluster_list.multiplicity_factors(num_sites_in_group)

    @property
    def all_cf_names(self):
        num_bf = len(self.basis_functions)
        return self.cluster_list.get_all_cf_names(num_bf)

    @property
    def num_cf(self):
        """Return the number of correlation functions."""
        return len(self.all_cf_names)

    def _get_symm_groups(self):
        symm_groups = -np.ones(len(self.atoms), dtype=int)

        for group, indices in enumerate(self.index_by_trans_symm):
            symm_groups[indices] = group
        return symm_groups.tolist()

    def _cutoff_for_tm_construction(self):
        """ Get cutoff radius for translation matrix construction."""
        indices = self.cluster_list.unique_indices
        # start with some small positive number
        max_dist = 0.1

        for ref in indices:
            # MIC distance is a lower bound for the distance used in the
            # cluster
            mic_distances = self.atoms.get_distances(ref, indices, mic=True)
            dist = np.max(mic_distances)
            if dist > max_dist:
                max_dist = dist

        # 0.5 * max_dist is the radius, but give a bit of buffer (0.1)
        max_dist *= 0.51
        return max_dist

    def create_cluster_list_and_trans_matrix(self):
        self._create_cluster_list()

        symm_group = self._get_symm_groups()
        tm_cutoff = self._cutoff_for_tm_construction()

        # For smaller cell we currently have no method to decide how large
        # cutoff we need in order to ensure that all indices in unique_indices
        # are included. We therefore just probe the cutoff and increase it by
        # a 1 angstrom until we achieve the what we want
        all_included = False
        counter = 0
        max_attempts = 1000
        supercell, ref_indices = self._get_supercell()

        # We need to get the symmetry groups of the supercell. We utilise that
        # the supercell is tagged
        symm_group_sc = [-1 for _ in range(len(supercell))]
        for atom in supercell:
            symm_group_sc[atom.index] = symm_group[atom.tag]

        # Make as efficient as possible by evaluating only a subset of
        # the indices
        indices = [-1 for _ in range(len(self.atoms))]
        for atom in supercell:
            if indices[atom.tag] == -1:
                indices[atom.tag] = atom.index

        unique_index_symm = self.cluster_list.unique_indices_per_group
        unique_index_symm = [set(x) for x in unique_index_symm]
        all_unique_indices = set(self.cluster_list.unique_indices)
        while not all_included and counter < max_attempts:
            try:
                tmc = TransMatrixConstructor(supercell, tm_cutoff)
                tm_sc = tmc.construct(ref_indices, symm_group_sc,
                                      indices=indices)

                # Map supercell indices to normal indices
                tm = trans_matrix_index2tags(tm_sc, supercell, indices=indices)
                for i, row in enumerate(tm):
                    _ = [row[k] for k in unique_index_symm[symm_group[i]]]

                # For a simpler data structure in calculator store some
                # additional data
                for i, row in enumerate(tm):
                    gr = symm_group[i]
                    diff = all_unique_indices - unique_index_symm[gr]
                    for i in diff:
                        row[i] = -1  # Put -1 as this should never be accessed
                all_included = True
            except (KeyError, IndexError):
                tm_cutoff += 3.0
            counter += 1

        if counter >= max_attempts:
            raise RuntimeError("Could not find a cutoff such that all "
                               "unique_indices are included")

        self.trans_matrix = [{k: int(row[k]) for k in
                              self.cluster_list.unique_indices} for row in tm]

    def _store_data(self):
        size_str = nested_list2str(self.size)
        num = self.template_atoms_uid
        num_templates = self.template_atoms.num_templates
        _logger('Generating cluster data for template with size: {}. '
                '({} of {})'.format(size_str, num+1, num_templates),
                verbose=LogVerbosity.INFO)

        db = connect(self.db_name)
        data = {'cluster_list': [x.todict()
                                 for x in self.cluster_list.tolist()],
                'trans_matrix': self.trans_matrix}

        try:
            row = db.get(name="template", size=self._size2string())
            db.update(row.id, data=data)
        except KeyError:
            db.write(self.atoms, name='template', data=data,
                     size=self._size2string())

    def _keys2int(self, tm):
        """
        Convert the keys in the translation matrix to integers
        """
        tm_int = []
        for row in tm:
            tm_int.append({int(k): v for k, v in row.items()})
        return tm_int

    def _read_data(self):
        db = connect(self.db_name)
        try:
            select_cond = [('name', '=', 'template'),
                           ('size', '=', self._size2string())]
            row = db.get(select_cond)
            info_str = row.data.cluster_list
            self.cluster_list.clear()
            for item in info_str:
                cluster = Cluster.load(item)
                self.cluster_list.append(cluster)
            self.trans_matrix = self._keys2int(row.data.trans_matrix)
        except KeyError:
            self.create_cluster_list_and_trans_matrix()
            self._store_data()
        except (AssertionError, AttributeError, RuntimeError):
            self.reconfigure_settings()

    def _group_index_by_basis(self):
        first_symb_in_basis = [x[0] for x in self.basis_elements]
        return self.atoms_mng.index_by_symbol(first_symb_in_basis)

    def _activate_lagest_template(self):
        atoms = self.template_atoms.largest_template_by_diag
        self.set_active_template(atoms)

    def view_clusters(self):
        """Display all clusters along with their names."""
        from ase.gui.gui import GUI
        from ase.gui.images import Images
        self._activate_lagest_template()
        figures = self.cluster_list.get_figures(self.atoms)
        images = Images()
        images.initialize(figures)
        gui = GUI(images, expr='')
        gui.show_name = True
        gui.run()

    def reconfigure_settings(self):
        """Reconfigure templates stored in DB file."""
        # Reconfigure the cluster information for each template based on
        # current max_cluster_size and max_cluster_dia
        for uid in range(self.template_atoms.num_templates):
            self._set_active_template_by_uid(uid)
            self.create_cluster_list_and_trans_matrix()
            self._store_data()
        self._set_active_template_by_uid(0)
        _logger('Cluster data updated for all templates.\n'
                'You should also reconfigure DB entries (in CorrFunction '
                'class) to make the information on each structure to be '
                'consistent.',
                verbose=LogVerbosity.INFO)

    def _check_first_elements(self):
        basis_elements = self.basis_elements
        num_basis = self.num_basis
        # This condition can be relaxed in the future
        first_elements = []
        for elements in basis_elements:
            first_elements.append(elements[0])
        if len(set(first_elements)) != num_basis:
            raise ValueError("First element of different basis should not be "
                             "the same.")

    def save(self, filename):
        """Write Setting object to a file in JSON format.

        Parameters:

        filename: str
            Name of the file to store the necessary settings to initialize
            the Cluster Expansion calculations.
        """
        class_types = ['CEBulk', 'CECrystal']
        if type(self).__name__ not in class_types:
            raise NotImplementedError('Class {} '
                                      'is not supported.'
                                      ''.format(type(self).__name__))

        import json
        if type(self).__name__ == 'CEBulk':
            self.kwargs['classtype'] = 'CEBulk'
        else:
            self.kwargs['classtype'] = 'CECrystal'
        # Write keyword arguments necessary for initializing the class
        with open(filename, 'w') as outfile:
            json.dump(self.kwargs, outfile, indent=2)

    def _check_cluster_list_consistency(self):
        """Check that cluster names in all templates' info entries match."""
        db = connect(self.db_name)
        ref_clust_list = None
        for row in db.select(name='template'):
            cluster_list_str = row.data['cluster_list']
            cluster_list = ClusterList()
            for item in cluster_list_str:
                cluster = Cluster.load(item)
                cluster_list.append(cluster)
            cluster_list.sort()

            if ref_clust_list is None:
                ref_clust_list = cluster_list

            assert cluster_list == ref_clust_list


def to_3x3_matrix(size):
    if size is None:
        return None
    if isinstance(size, np.ndarray):
        size = size.tolist()

    # Is already a matrix
    if np.array(size).shape == (3, 3):
        return size

    if np.array(size).shape == (3,):
        return np.diag(size).tolist()

    raise ValueError("Cannot convert passed array {} to 3x3 matrix"
                     "".format(size))



class MCConstraint(object):
    """
    Class for that prevents the MC sampler to run certain moves
    """

    def __init__(self):
        self.name = "GenericConstraint"

    def __call__(self, system_changes):
        """Return `True` if the trial move is valid.

        Parameters:

        system_changes: list
            System changes. See doc-string of
            `clease.montecarlo.observers.MCObserver`
        """
        return True



import json
import os
from clease import CorrFunction
from clease.calculator import Clease
from clease.tools import nested_list2str


def attach_calculator(setting=None, atoms=None, eci={},
                      fname_prefix=None, load=True):
    """Utility function for an efficient initialization of large cells.

    Parameters:

    setting: `ClusterExpansionSetting` object

    eci: dict
        Dictionary containing cluster names and their ECI values

    atoms: Atoms object
        Atoms object for MC simulations.

    fname_prefix: str
        Prefix of the file name for backing up cluster info in .json format.
        The file name will append '_cluster_info_{size_of_the_cell}.json'.

    load: bool
        Load cluster info if possible (the file with a name specified using
        the fname_prefix is present).
    """
    cf_names = list(eci.keys())
    init_cf = CorrFunction(setting).get_cf_by_names(setting.atoms, cf_names)

    template_uid = setting.template_atoms.get_uid_matching_atoms(
        atoms=atoms, generate_template=True)
    setting.prepare_new_active_template(template_uid)

    size_str = nested_list2str(setting.size)

    fname = '.'
    if fname_prefix is not None:
        fname = fname_prefix + '_cluster_info_{}.json'.format(size_str)

    loaded_info = False
    if load and os.path.exists(fname) and fname_prefix is not None:
        with open(fname, 'r') as infile:
            data = json.load(infile)
        setting.cluster_list = data['cluster_list']
        setting.trans_matrix = data['trans_matrix']
        loaded_info = True
    else:
        setting.create_cluster_list_and_trans_matrix()

    data = {'cluster_list': setting.cluster_list,
            'trans_matrix': setting.trans_matrix,
            'size': setting.size,
            'setting': setting.kwargs}

    if fname_prefix is not None and not loaded_info:
        save_info(fname, data)

    atoms = setting.atoms.copy()

    calc = Clease(setting, eci=eci, init_cf=init_cf)
    atoms.set_calculator(calc)
    return atoms


def save_info(fname, data):
    with open(fname, 'w') as outfile:
        json.dump(data, outfile)



import numpy as np
from scipy.spatial import cKDTree as KDTree
from itertools import combinations
from clease.cluster_fingerprint import ClusterFingerprint


class ClusterExtractor(object):
    """
    Extract clusters that contain the suppied atomic index.


    Parameters:

    atoms: Atoms object
        ASE Atoms object
    """

    def __init__(self, atoms):
        self.atoms = atoms
        self.tree = KDTree(self.atoms.get_positions())
        self.inner_prod = []
        self.tol = 1E-6

    def extract(self, ref_indx=0, size=2, cutoff=4.0, ignored_indices=[]):
        """
        Extract all clusters with a given size if they are smaller than the
        cutoff distance.

        This method returns a list of sites (atomic indices) belonging to
        every cluster. If there are *N* clusters, it returns a list of a form
            [cluster_1, cluster_2, cluster_3, cluster_N]
        where cluster_x is a nested list in a form
            cluster1 = [[245, 432, 126], [567, 432, 127], ...]


        Parameters:

        ref_indx: int
            Index to be included in all of the clusters

        size: int
            Cluster size (i.e., number of atoms in a cluster)

        cutoff: float
            Maximum cutoff distance

        ignored_indices: list
            all of the background indices to be ignored when creating clusters
        """
        self.inner_prod = []
        x = self.atoms.get_positions()[ref_indx, :]
        indices = self.tree.query_ball_point(x, cutoff)
        indices.remove(ref_indx)
        indices = list(set(indices) - set(ignored_indices))
        return self._group_clusters(ref_indx, indices, size, cutoff)

    def _get_type(self, fingerprint):
        """Determine cluster type based on flattened inner product matrix."""
        if self.inner_prod:
            try:
                group = self.inner_prod.index(fingerprint)
                return group
            except ValueError:
                pass

        self.inner_prod.append(fingerprint)
        return len(self.inner_prod) - 1

    def _group_clusters(self, ref_indx, indices, size, cutoff):
        """Group sites in clusters based on the inner product matrix."""
        pos = self.atoms.get_positions()
        clusters = []
        for comb in combinations(indices, r=size-1):
            all_indices = [ref_indx] + list(comb)

            d = self.get_max_distance(all_indices)
            if d > cutoff:
                continue
            X = pos[all_indices, :]
            X -= np.mean(X, axis=0)
            X = X.dot(X.T)

            diag = np.diagonal(X)
            off_diag = []
            for i in range(1, X.shape[0]):
                off_diag += np.diagonal(X, offset=i).tolist()
            N = X.shape[0]
            assert len(off_diag) == N*(N-1)/2

            inner = np.array(sorted(diag, reverse=True) + sorted(off_diag))
            fp = ClusterFingerprint(list(inner))
            group = self._get_type(fp)

            all_indices = self._order_by_internal_distances(all_indices)
            if group == len(clusters):
                clusters.append([all_indices])
            else:
                clusters[group].append(all_indices)
        return clusters

    def _order_by_internal_distances(self, clusters):
        """Order the indices by internal distances."""
        dists = self._get_internal_distances(clusters)
        zipped = sorted(list(zip(dists, clusters)), reverse=True)
        return [x[1] for x in zipped]

    def view_figures(self, cluster):
        """Visualize figures of a cluster."""
        from ase.visualize import view
        figures = [self.atoms[x] for x in cluster]
        view(figures)

    def _get_internal_distances(self, figure):
        """Calculate all internal distances of a figure."""
        dists = []
        for indx in figure:
            d = self.atoms.get_distances(indx, figure)
            d = d.round(decimals=6)
            dists.append(sorted(d.tolist(), reverse=True))
        return dists

    def equivalent_sites(self, figure):
        """Find the equivalent sites of a figure."""
        dists = self._get_internal_distances(figure)
        equiv_sites = []
        for i in range(len(dists)):
            for j in range(i+1, len(dists)):
                if np.allclose(dists[i], dists[j]):
                    equiv_sites.append((i, j))

        # Merge pairs into groups
        merged = []
        for equiv in equiv_sites:
            found_group = False
            for m in merged:
                if any(x in m for x in equiv):
                    m.update(equiv)
                    found_group = True
            if not found_group:
                merged.append(set(equiv))
        return [list(x) for x in merged]

    def get_max_distance(self, figure):
        """Return the maximum distance of a figure."""
        internal_dists = self._get_internal_distances(figure)
        max_dists = [x[0] for x in internal_dists]
        return max(max_dists)



from clease.montecarlo.observers import MCObserver


class ConcentrationObserver(MCObserver):
    """
    Observer that can be attached to a MC run, to track the concenctration of a
    particular element.

    Parameters:

    atoms: Atoms object
        Atoms object used for MC

    element: str
        The element that should be tracked
    """
    def __init__(self, atoms, element=None):
        self.element = element
        self.n = len(atoms)
        self.init_conc = self.calculate_from_scratch(atoms)
        self.current_conc = self.init_conc
        self.avg_conc = self.current_conc
        self.avg_conc_sq = self.current_conc**2
        self.num_calls = 1

    def new_concentration(self, system_change):
        """Calculate the new consentration after the changes."""
        new_conc = self.current_conc
        for change in system_change:
            if change[2] == self.element:
                new_conc += 1.0/self.n
            if change[1] == self.element:
                new_conc -= 1.0/self.n
        return new_conc

    def __call__(self, system_change, peak=False):
        if system_change is None:
            return self.current_conc

        new_conc = self.new_concentration(system_change)

        if peak:
            return new_conc

        self.current_conc = new_conc
        self.avg_conc += self.current_conc
        self.avg_conc_sq += self.current_conc**2
        self.num_calls += 1
        return self.current_conc

    def reset(self):
        """Reset the averages."""
        self.avg_conc = self.current_conc
        self.avg_conc_sq = self.current_conc**2
        self.num_calls = 1

    def get_averages(self):
        mean_conc = self.avg_conc/self.num_calls
        var_conc = self.avg_conc_sq/self.num_calls - mean_conc**2
        return {
            'conc_{}'.format(self.element): mean_conc,
            'conc_var_{}'.format(self.element): var_conc
        }

    def calculate_from_scratch(self, atoms):
        num_atoms = sum(1 for a in atoms if a.symbol == self.element)
        return num_atoms/len(atoms)



# from kivy.uix.floatlayout import FloatLayout
from kivy.uix.stacklayout import StackLayout
from kivy.properties import ObjectProperty


class LoadDialog(StackLayout):
    load = ObjectProperty(None)
    cancel = ObjectProperty(None)


class SaveDialog(StackLayout):
    save = ObjectProperty(None)
    cancel = ObjectProperty(None)



from clease.montecarlo.bias_potential import BiasPotential
from clease.montecarlo.montecarlo import Montecarlo
from clease.montecarlo.sgc_montecarlo import SGCMonteCarlo
from clease.montecarlo.gaussian_kernel_bias_potential import GaussianKernelBiasPotential
from clease.montecarlo.binned_bias_potential import BinnedBiasPotential
from clease.montecarlo.metadynamics_sampler import MetaDynamicsSampler

__all__ = ['Montecarlo']



from kivy.app import App
from kivy.lang import Builder
from kivy.uix.stacklayout import StackLayout
from kivy.resources import resource_add_path
from kivy.uix.popup import Popup

from clease.gui.settingsPage import SettingsPage
from clease.gui.concentrationPage import ConcentrationPage
from clease.gui.newStructPage import NewStructPage
from clease.gui.fitPage import FitPage
from clease.gui.meta_dyn_page import MetaDynPage
from clease.gui.reconfigDB import ReconfigDB
from kivy.core.window import Window
from clease.gui.job_exec import JobExec
from clease.gui.mc_header import MCHeader
from clease.gui.load_save_dialog import LoadDialog, SaveDialog
from clease.gui.db_browser import DbBrowser
from ase.db import connect
import subprocess
import signal
from pathlib import Path
import traceback

try:
    import ase.db.app as ase_db_webapp
    from flask import request
    has_flask = True
except ImportError:
    has_flask = False

import json
from threading import Thread

import os.path as op
import os

main_path = Path(__file__)
resource_add_path(main_path.parent / 'layout')

Builder.load_file("cleaseGUILayout.kv")


class WindowFrame(StackLayout):
    _pop_up = None
    current_session_file = None
    settings = None
    reconfig_in_progress = False
    subprocesses = {}
    active_template_is_mc_cell = False
    info_update_disabled = False
    view_mc_cell_disabled = False

    def dismiss_popup(self):
        if self._pop_up is None:
            return
        self._pop_up.dismiss()
        self._pop_up = None

    def show_load_session_dialog(self):
        content = LoadDialog(load=self.load_session, cancel=self.dismiss_popup)
        self._pop_up = Popup(title="Load CLEASE session", content=content,
                             pos_hint={'right': 0.95, 'top': 0.95},
                             size_hint=(0.9, 0.9))
        self._pop_up.open()

    def load_session(self, path, filename):
        try:
            with open(filename[0], 'r') as infile:
                data = json.load(infile)

            # variables for "Concentration" screen
            conc_page = self.ids.sm.get_screen('Concentration')
            elements = data['conc']['elements']
            grouped_basis = data['conc']['grouped_basis']
            conc_page.set_Elements_GroupedBasis(elements, grouped_basis)

            A_lb = data['conc']['A_lb']
            rhs_lb = data['conc']['rhs_lb']
            A_eq = data['conc']['A_eq']
            rhs_eq = data['conc']['rhs_eq']
            conc_page.load_from_matrices(A_lb, rhs_lb, A_eq, rhs_eq)

            # variables for "Settings" screen
            settings_page = self.ids.sm.get_screen('Settings')
            settings_page.from_dict(data.get('settings', {}))
            settings_page.apply_settings()

            newstruct_page = self.ids.sm.get_screen('NewStruct')
            newstruct_page.from_dict(data.get('new_struct', {}))

            job_exec = self.ids.sm.get_screen('JobExec')
            job_exec.from_dict(data.get('job_exec', {}))

            fit_page = self.ids.sm.get_screen('Fit')
            fit_page.from_dict(data.get('fit_page', {}))

            mc_header = self.ids.sm.get_screen('MCHeader')
            mc_main_page = mc_header.ids.sm.get_screen('MCMainPage')
            mc_main_page.from_dict(data.get('mc_main', {}))
            canonical_page = mc_header.ids.sm.get_screen('MC')
            canonical_page.from_dict(data.get('canonical_mc', {}))

            meta_dyn_page = mc_header.ids.sm.get_screen('MetaDynPage')
            meta_dyn_page.from_dict(data.get('meta_dyn_page', {}))

            self.current_session_file = filename[0]

            msg = "Loaded session from {}".format(self.current_session_file)
            App.get_running_app().root.ids.status.text = msg

        except Exception as e:
            msg = "An error occured during load: " + str(e)
            App.get_running_app().root.ids.status.text = msg

        self.dismiss_popup()

    def save_session_to_current_file(self):
        if self.current_session_file is not None:
            self.save_session(None, [self.current_session_file], None)

        else:
            self.show_save_dialog()

    def save_session(self, path, selection, user_filename):
        if self.ids.sm.get_screen('Settings').check_user_input() != 0:
            return

        if self.ids.sm.get_screen('Concentration').check_user_input() != 0:
            return

        if len(selection) == 0:
            fname = str(Path(path) / user_filename)
        else:
            fname = selection[0]

        data = {}
        data['settings'] = self.ids.sm.get_screen('Settings').to_dict()
        data['conc'] = self.ids.sm.get_screen('Concentration').to_dict()
        data['new_struct'] = self.ids.sm.get_screen('NewStruct').to_dict()
        data['job_exec'] = self.ids.sm.get_screen('JobExec').to_dict()
        data['fit_page'] = self.ids.sm.get_screen('Fit').to_dict()
        mc_header = self.ids.sm.get_screen('MCHeader')
        data['mc_main'] = mc_header.ids.sm.get_screen('MCMainPage').to_dict()
        data['canonical_mc'] = mc_header.ids.sm.get_screen('MC').to_dict()
        data['meta_dyn_page'] = mc_header.ids.sm.get_screen('MetaDynPage').to_dict()

        with open(fname, 'w') as outfile:
            json.dump(data, outfile, separators=(',', ': '), indent=2)
        msg = 'Session saved to {}'.format(fname)
        App.get_running_app().root.ids.status.text = msg
        self.dismiss_popup()
        self.current_session_file = fname

    def show_save_dialog(self):
        if self.ids.sm.get_screen('Settings').check_user_input() != 0:
            return

        if self.ids.sm.get_screen('Concentration').check_user_input() != 0:
            return

        content = SaveDialog(save=self.save_session, cancel=self.dismiss_popup)

        self._pop_up = Popup(title="Save CLEASE session", content=content,
                             pos_hint={'right': 0.95, 'top': 0.95},
                             size_hint=(0.9, 0.9))
        self._pop_up.open()

    def change_screen(self, new_screen):
        current = self.ids.sm.current
        all_screens = self.ids.sm.screen_names

        index_current = all_screens.index(current)
        index_new = all_screens.index(new_screen)

        direction = 'left'

        if index_current > index_new:
            direction = 'right'
        self.ids.sm.transition.direction = direction

        # Take special care of MC screens as
        if new_screen == 'MC':
            new_screen = self.mc_type_screen
        self.ids.sm.current = new_screen

    def reconfig(self, target=None):
        """Reconfigure target.

        Parameters:

        target: str
            one of "settings", "db" and "settings_db"
        """
        if self.reconfig_in_progress:
            # Do no allow user to initialize many threads
            return

        self.reconfig_in_progress = True

        reconfig = ReconfigDB()
        reconfig.app = App.get_running_app()
        reconfig.status = App.get_running_app().root.ids.status

        if target == 'settings':
            Thread(target=reconfig.reconfig_settings).start()
        elif target == 'db':
            Thread(target=reconfig.reconfig_db).start()
        else:
            Thread(target=reconfig.reconfig_settings_db).start()

    def view_clusters(self):
        """View clusters."""
        images = self._get_clusters()

        if images is False:
            msg = "Settings should be applied/loaded before viewing clusters."
            App.get_running_app().root.ids.status.text = str(msg)
            return

        try:
            from ase.visualize import view
            Thread(target=view, args=(images,)).start()

        except Exception as exc:
            App.get_running_app().root.ids.status.text = str(exc)

    def _get_clusters(self):
        if self.settings is None:
            self._apply_settings()

        try:
            self.settings._activate_lagest_template()
            atoms = self.settings.atoms
            return self.settings.cluster_list.get_figures(atoms)
        except Exception:
            return False

    def _apply_settings(self):
        self.ids.sm.get_screen('Settings').apply_settings()

    def view_training_db(self):
        if self.settings is None:
            self._apply_settings()

        app = App.get_running_app()
        if self.settings is None:
            msg = 'Could not apply settings. Check your input.'
            app.root.ids.status.text = msg
            return

        screen = self.ids.sm.get_screen('Settings')
        db_name = screen.ids.dbNameInput.text

        content = DbBrowser(close=self.dismiss_popup, db_name=db_name)
        self._pop_up = Popup(title="DB Browser", content=content,
                             pos_hint={'right': 0.95, 'top': 0.95},
                             size_hint=(0.9, 0.9))
        self._pop_up.open()

    def get_mc_cell(self):
        app = App.get_running_app()
        try:
            settings = app.root.settings

            mc_header = self.ids.sm.get_screen('MCHeader')

            mc_main_page = mc_header.ids.sm.get_screen('MCMainPage') 

            size = int(mc_main_page.ids.sizeInput.text)

            if settings is None:
                app.root.ids.status.text = 'Apply settings prior to running MC'
                return

            atoms = None
            if self.active_template_is_mc_cell:
                atoms = settings.atoms.copy()
            else:
                atoms = settings.atoms*(size, size, size)
            return atoms
        except Exception as exc:
            traceback.print_exc()
            app.root.ids.status.text = str(exc)
        return None

    def view_mc_cell(self):
        app = App.get_running_app()
        if self.view_mc_cell_disabled:
            msg = 'Cannot view MC cell while attaching calculator'
            app.root.ids.status.text = msg
            return
        try:
            from ase.visualize import view
            atoms = self.get_mc_cell()
            Thread(target=view, args=(atoms,)).start()
        except Exception as exc:
            traceback.print_exc()
            app.root.ids.status.text = str(exc)



class CleaseGUI(App):
    def __init__(self):
        App.__init__(self)
        self.settings = None

    def build(self):
        self.icon = 'clease_logo.png'
        Window.bind(on_keyboard=self.on_keyboard)
        return WindowFrame()

    def on_keyboard(self, window, key, scancode, codepoint, modifier):
        if modifier in[['ctrl'], ['meta']] and codepoint == 's':
            self.root.save_session_to_current_file()

    def on_stop(self):
        for k, v in self.root.subprocesses.items():
            os.kill(v, signal.SIGTERM)
        self.root.subprocesses = {}


if __name__ == "__main__":
    CleaseGUI().run()



"""Class containing a manager for creating template atoms."""
import os
import numpy as np
from itertools import product
from random import choice, shuffle
from ase.db import connect
from ase.build import cut
from itertools import combinations
from ase.build import make_supercell
from clease.tools import str2nested_list
from clease import _logger
from clease import SkewnessFilter, EquivalentCellsFilter
from clease.template_filters import CellFilter, AtomsFilter
from clease.tools import all_integer_transform_matrices


class TemplateAtoms(object):
    def __init__(self, supercell_factor=None, size=None, skew_threshold=4,
                 db_name=None):
        if size is None and supercell_factor is None:
            raise TypeError("Either size or supercell_factor needs to be "
                            "specified.\n size: list or numpy array.\n "
                            "supercell_factor: int")

        self.supercell_factor = supercell_factor
        self.size = size

        self.cell_filters = []
        self.atoms_filters = []

        self.add_cell_filter(SkewnessFilter(skew_threshold))
        self.all_cells = []
        self.add_cell_filter(EquivalentCellsFilter(self.all_cells))

        if self.size is not None:
            check_valid_conversion_matrix(self.size)
        self.skew_threshold = skew_threshold
        self.templates = None
        self.db_name = db_name
        self.db = connect(db_name)
        self.prim_cell = \
            list(self.db.select(name='primitive_cell'))[0].toatoms()
        self._set_based_on_setting()
        self._append_templates_from_db()

    def __str__(self):
        """Print a summary of the class."""
        msg = "=== TemplateAtoms ===\n"
        msg += "Supercell factor: {}\n".format(self.supercell_factor)
        msg += "Skewed threshold: {}\n".format(self.skew_threshold)
        msg += "Template sizes:\n"
        for size in self.templates['size']:
            msg += "{}\n".format(size)
        return msg

    @property
    def num_templates(self):
        return len(self.templates['atoms'])

    def add_cell_filter(self, cell_filter):
        """Attach a new Cell filter."""
        if not isinstance(cell_filter, CellFilter):
            raise TypeError("filter has to be an instance of CellFilter")
        self.cell_filters.append(cell_filter)

    def add_atoms_filter(self, at_filter):
        """Attach a new Atoms filter."""
        if not isinstance(at_filter, AtomsFilter):
            raise TypeError("filter has to be an instance of AtomsFilter")
        self.atoms_filters.append(at_filter)

    def clear_filters(self):
        """Remove all filters."""
        self.cell_filters = []
        self.atoms_filters = []

    def remove_filter(self, f):
        """Remove one filter"""
        if isinstance(f, AtomsFilter):
            self.atoms_filters.remove(f)
        elif isinstance(f, CellFilter):
            self.cell_filters.remove(f)
        else:
            raise TypeError('Only AtomsFilters and CellFilters '
                            'can be removed')

    def remove_filters(self, filters):
        """Remove a list of filters."""
        for f in filters:
            self.remove_filter(f)

    def apply_filter(self, template_filter):
        """Apply a filter to already generated templates.

        Paramaters:

        template_filter: AtomsFilter or CellFilter
            Filter to run through the templates
        """
        filtered_templates = {
            'size': [],
            'atoms': []
        }
        for size, at in zip(self.templates['size'], self.templates['atoms']):
            if template_filter(at):
                filtered_templates['size'].append(size)
                filtered_templates['atoms'].append(at)
        self.templates = filtered_templates

    def is_valid(self, atoms=None, cell=None):
        """
        Check the validity of the template.

        Return `True` if templates are valid according to the attached filters.

        Parameters:

        atoms: Atoms object

        cell: unit cell vector
        """
        if atoms is None and cell is None:
            msg = "At least one of `atoms` or `cell` must be specified."
            raise ValueError(msg)

        cell_valid = True
        if cell is not None:
            cell_valid = all([filter(cell) for filter in self.cell_filters])

        if not cell_valid:
            return False

        atoms_valid = True
        if atoms is not None:
            atoms_valid = all([f(atoms) for f in self.atoms_filters])

        return cell_valid and atoms_valid

    def get_size(self):
        """Get size of the templates."""
        return self.templates['size']

    def get_atoms(self, uid, return_size=False):
        """Return atoms at position."""
        if return_size:
            assert is_3x3_matrix(self.templates['size'][uid])
            return self.templates['atoms'][uid], self.templates['size'][uid]

        return self.templates['atoms'][uid]

    @property
    def largest_template_by_num_atom(self):
        """Return the largest template based on number of atoms it has."""
        max_num = 0
        largest_template = None
        for atoms in self.templates['atoms']:
            if len(atoms) > max_num:
                largest_template = atoms
                max_num = len(atoms)
        return largest_template

    @property
    def largest_template_by_diag(self):
        """Return the largest template based on the shortest diagonal."""
        length = 0.0
        largest_template = None
        for atoms in self.templates['atoms']:
            diag_lengths = []
            cell = atoms.get_cell().T
            for w in product([-1, 0, 1], repeat=3):
                if np.allclose(w, 0):
                    continue
                diag = cell.dot(w)
                length = np.sqrt(diag.dot(diag))
                diag_lengths.append(length)

            min_length = np.min(diag_lengths)

            if min_length > length:
                largest_template = atoms
                length = min_length

        return largest_template

    def get_uid_with_given_size(self, size, generate_template=False):
        """Get the UID of the template with given size.

        Parameters:

        size: list of length 3

        generate_template: bool (optional)
            If *True*, generate a new template if a template with matching
            size is not found.
        """
        if not is_3x3_matrix(size):
            raise ValueError("Expect the size to be a list of 3x3 matrix")

        try:
            uid = self.templates['size'].index(size)
            return uid
        except IndexError:
            pass

        if not generate_template:
            raise ValueError("There is no template with size = {}."
                             "".format(size))

        # get dims based on the passed atoms and append.
        _logger("Template that matches the specified size not found. "
                "Generating...")
        check_valid_conversion_matrix(size)
        prim_cell = self.prim_cell
        self.templates['atoms'].append(prim_cell*size)
        self.templates['size'].append(size)
        self._check_templates_datastructure()

        return len(self.templates['atoms']) - 1

    def get_uid_matching_atoms(self, atoms, generate_template=False):
        """Get the UID for the template matching atoms.

        Parameters:

        atoms: Atoms object
            structure to compare its size against template atoms

        generate_template: bool (optional)
            If *True*, generate a new template if a template with matching
            size is not found.
        """
        shape = atoms.get_cell_lengths_and_angles()
        for uid, template in enumerate(self.templates['atoms']):
            shape_template = template.get_cell_lengths_and_angles()
            if np.allclose(shape, shape_template):
                return uid

        if not generate_template:
            raise RuntimeError("There is no template that matches the shape "
                               "of given atoms object")

        # get dims based on the passed atoms and append.
        _logger("Template that matches the size of passed atoms not found. "
                "Generating...")
        size = self._get_conversion_matrix(atoms)
        assert is_3x3_matrix(size)

        prim_cell = self.prim_cell

        self.templates['atoms'].append(make_supercell(prim_cell, size))
        self.templates['size'].append(list(size))
        self._check_templates_datastructure()
        return len(self.templates['atoms']) - 1

    def _set_based_on_setting(self):
        """Construct templates based on arguments specified."""
        if self.size is None:
            self.supercell_factor = int(self.supercell_factor)
            self.templates = self._generate_template_atoms()
            if not self.templates['atoms']:
                raise RuntimeError("No template atoms with matching criteria")
        else:
            assert is_3x3_matrix(self.size)
            # if size and supercell_factor are both specified,
            # size will be used
            prim_cell = self.prim_cell
            self.templates = {'atoms': [make_supercell(prim_cell, self.size)],
                              'size': [self.size]}
        self._check_templates_datastructure()

    def _append_templates_from_db(self):
        if not os.path.isfile(self.db_name):
            return
        for row in self.db.select(name='template'):
            found = False
            size_str = row.get('size', None)
            if size_str is None:
                continue
            for i, _ in enumerate(self.templates['atoms']):
                size = str2nested_list(size_str)
                assert is_3x3_matrix(size)
                if (self.templates['size'][i] == size):
                    found = True
                    break

            if not found:
                atoms = make_supercell(self.prim_cell, size)
                self.templates['atoms'].append(atoms)
                self.templates['size'].append(size)
        self._check_templates_datastructure()

    def _generate_template_atoms(self):
        """Generate all template atoms up to a certain multiplicity factor."""
        templates = {'atoms': [], 'size': []}
        # case 1: size of the cell is given
        if self.size is not None:
            assert is_3x3_matrix(self.size)

            atoms = make_supercell(self.prim_cell, self.size)
            templates['atoms'].append(atoms)
            templates['size'].append(self.size)
            return templates

        # case 2: supercell_factor is given
        for size in product(range(1, self.supercell_factor+1), repeat=3):
            # Skip cases where the product of factors is larger than the
            # supercell factor.
            if np.prod(size) > self.supercell_factor:
                continue
            matrix = np.diag(size)
            atoms = make_supercell(self.prim_cell, matrix)

            if self.is_valid(atoms=atoms, cell=atoms.get_cell()):
                templates['atoms'].append(atoms)
                templates['size'].append(matrix.tolist())
                self.all_cells.append(atoms.get_cell())
        return templates

    def _construct_templates_from_supercell(self, size):
        """Construct subcells based on a supercell."""
        templates = {'atoms': [], 'size': []}

        # Select the first unit cell entry in the DB
        ucell = self.prim_cell
        V = ucell.get_volume()

        supercell = ucell*size
        com = supercell.get_center_of_mass()

        dists = supercell.get_positions() - com
        lengths_squared = np.sum(dists**2, axis=1)
        closest = np.argmin(lengths_squared)
        pos = supercell.get_positions()
        origo = pos[closest, :]
        pos -= origo

        # Construct all templates
        symb = supercell[closest].symbol
        indices = [atom.index for atom in supercell if atom.symbol == symb]
        indices.remove(closest)

        distances = supercell.get_distances(closest, indices)
        sorted_indices = np.argsort(distances).tolist()
        indices = [indices[i] for i in sorted_indices]

        inv_cell = np.linalg.inv(ucell.get_cell().T)

        scaled_pos = inv_cell.dot(pos.T).T
        filtered_indices = []
        for indx in indices:
            vec = np.round(scaled_pos[indx, :])
            if np.allclose(vec.astype(np.int32), vec):
                filtered_indices.append(indx)
        indices = filtered_indices

        for i1, i2, i3 in combinations(indices, r=3):
            if i1 == i2 or i1 == i3 or i2 == i3:
                continue

            try:
                v1 = pos[i1, :]
                v2 = pos[i2, :]
                v3 = pos[i3, :]
                v1 = inv_cell.dot(v1)
                v2 = inv_cell.dot(v2)
                v3 = inv_cell.dot(v3)
                atoms = cut(ucell, a=v1, b=v2, c=v3)
            except np.linalg.linalg.LinAlgError:
                # Can't construct a 3D cell
                continue
            new_vol = atoms.get_volume()

            if new_vol > V*self.supercell_factor:
                continue

            if not self.is_valid(atoms=atoms, cell=atoms.get_cell()):
                continue

            size = [[int(np.round(x)) for x in v1.tolist()],
                    [int(np.round(x)) for x in v2.tolist()],
                    [int(np.round(x)) for x in v3.tolist()]]
            templates['atoms'].append(atoms)
            templates['size'].append(size)
        return templates

    def _get_conversion_matrix(self, atoms):
        """Return the conversion matrix factor."""
        prim_cell = self.prim_cell

        small_cell = prim_cell.get_cell()
        inv_cell = np.linalg.inv(small_cell)

        large_cell = atoms.get_cell()
        size_factor = large_cell.dot(inv_cell)
        scale_int = size_factor.round(decimals=0).astype(int)
        if np.allclose(size_factor, scale_int):
            check_valid_conversion_matrix(scale_int)
            return scale_int.tolist()

        raise ValueError("The passed atoms object cannot be described by "
                         "repeating of the unit cells. Scale factors found "
                         "{}".format(size_factor))

    def _internal_distances_are_equal(self, atoms1, atoms2):
        """Check if all internal distances are equivalent."""
        if len(atoms1) != len(atoms2):
            return False
        dist1 = []
        dist2 = []
        for ref in range(len(atoms1)-1):
            remaining = range(0, len(atoms1))
            dist1 += atoms1.get_distances(ref, remaining).tolist()
            dist2 += atoms2.get_distances(ref, remaining).tolist()
        return np.allclose(sorted(dist1), sorted(dist2))

    def random_template(self, max_supercell_factor=1000):
        """Select a random template atoms.

        Parameters:

        max_supercell_factor: int
            Maximum supercell_factor the returned object can have
        """
        found = False
        num = 0
        while not found:
            num = choice(range(len(self.templates['atoms'])))
            factor = np.prod(self.templates['size'][num])
            if factor <= max_supercell_factor:
                found = True
        return num

    def _get_max_min_diag_ratio(self, atoms):
        """Return the ratio between the maximum and the minimum diagonal."""
        diag_lengths = []
        cell = atoms.get_cell().T
        for w in product([-1, 0, 1], repeat=3):
            if np.allclose(w, 0):
                continue
            diag = cell.dot(w)
            length = np.sqrt(diag.dot(diag))
            diag_lengths.append(length)
        max_length = np.max(diag_lengths)
        min_length = np.min(diag_lengths)
        return max_length/min_length

    def weighted_random_template(self):
        """Select a random template atoms with a bias towards a cubic cell.

        The bias is towards cells that have similar values for x-, y- and
        z-dimension sizes.
        """
        p_select = []
        for atoms in self.templates['atoms']:
            ratio = self._get_max_min_diag_ratio(atoms)
            p = np.exp(-4.0*(ratio-1.0)/self.skew_threshold)
            p_select.append(p)
        p_select = np.array(p_select)
        p_select /= np.sum(p_select)

        cum_prob = np.cumsum(p_select)
        rand_num = np.random.rand()
        indx = np.argmax(cum_prob > rand_num)
        return indx

    def _check_templates_datastructure(self):
        """Fails if the datastructure is inconsistent."""
        num_entries = len(self.templates['atoms'])
        for _, v in self.templates.items():
            assert len(v) == num_entries

    def has_atoms_filters(self):
        return len(self.atoms_filters) > 0

    def get_fixed_volume_templates(self, num_prim_cells=10, num_templates=10):
        # Set up a filter that listens to the templates with fixed volume
        from ase.build.tools import niggli_reduce_cell
        cells = []
        transform_matrices = []
        prim_vol = self.prim_cell.get_volume()
        for atoms in self.templates['atoms']:
            if abs(atoms.get_volume() - num_prim_cells*prim_vol) < 1E-6:
                cells.append(atoms.get_cell())
                transform_matrices.append(self._get_conversion_matrix(atoms))

        equiv_filter = EquivalentCellsFilter(cells)
        self.add_cell_filter(equiv_filter)

        ucell = self.prim_cell.get_cell()
        matrices = list(all_integer_transform_matrices(num_prim_cells))
        shuffle(matrices)
        for mat in matrices:
            sc = mat.dot(ucell)
            sc, _ = niggli_reduce_cell(sc)
            if self.is_valid(cell=sc):
                # If Atoms filters are present we check if it is valid
                at_valid = True
                if self.has_atoms_filters():
                    atoms = make_supercell(self.prim_cell, mat)
                    at_valid = self.is_valid(atoms=atoms)
                if at_valid:
                    cells.append(sc)
                    transform_matrices.append(mat)

            if len(transform_matrices) >= num_templates:
                break

        templates = []

        if len(transform_matrices) > num_templates:
            shuffle(transform_matrices)
            transform_matrices = transform_matrices[:num_templates]

        for P in transform_matrices:
            atoms = make_supercell(self.prim_cell, P)
            templates.append(atoms)

        # Remove the filter that was artificially added
        self.cell_filters.remove(equiv_filter)
        return templates


def is_3x3_matrix(array):
    return np.array(array).shape == (3, 3)


def check_valid_conversion_matrix(array):
    """
    Make sure that we have a right-handed coordinate system.
    Raise a ValueError if the matrix is not valid.
    """
    determinant = np.linalg.det(array)
    if determinant < 0.0:
        raise ValueError("The determinant of the size matrix is less than "
                         "zero (got {}). For a right "
                         "coordinate system, we need a positive "
                         "determinant!".format(determinant))



"""Module that fits ECIs to energy data."""
import os
import sys
import numpy as np
import multiprocessing as mp
import logging as lg
from clease import _logger
from ase.utils import basestring
from clease import CEBulk, CECrystal
from clease.mp_logger import MultiprocessHandler
from ase.db import connect


# Initialize a module-wide logger
logger = lg.getLogger(__name__)
logger.setLevel(lg.INFO)


class Evaluate(object):
    """Evaluate RMSE/MAE of the fit and CV scores.

    Parameters:

    setting: CEBulk or BulkSapcegroup object

    cf_names: list
        Names of clusters to include in the evalutation.
        If None, all of the possible clusters are included.

    select_cond: tuple or list of tuples (optional)
        Custom selection condition specified by user.
        Default only includes "converged=True" and
        "struct_type='initial'".

    max_cluster_size: int
        Maximum number of atoms in the cluster to include in the fit.
        If this is None then all clusters in the DB are included.

    max_cluster_size: int
        maximum number of atoms in the cluster to include in the fit.
        If ``None``, no restriction on the number of atoms will be imposed.

    max_cluster_dia: float or int
        maximum diameter of the cluster (in angstrom) to include in the fit.
        If ``None``, no restriction on the diameter.

    scoring_scheme: str
        should be one of 'loocv', 'loocv_fast' or 'k-fold'

    min_weight: float
        Weight given to the data point furthest away from any structure on the
        convex hull. An exponential weighting function is used and the decay
        rate is calculated as

        decay = log(min_weight)/min(sim_measure)

        where sim_measure is a similarity measure used to asses how different
        the structure is from structures on the convex hull.

    nsplits: int
        Number of splits to use when partitioning the dataset into
        training and validation data. Only used when scoring_scheme='k-fold'

    num_repetitions: int
        Number of repetitions used to use when calculating k-fold cross
        validation. The partitioning is repeated num_repetitions times
        and the resulting value is the average of the k-fold cross
        validation score obtained in each of the runs.
    """

    def __init__(self, setting, cf_names=None, select_cond=None,
                 parallel=False, num_core="all", fitting_scheme="ridge",
                 alpha=1E-5, max_cluster_size=None, max_cluster_dia=None,
                 scoring_scheme='loocv', min_weight=1.0, nsplits=10,
                 num_repetitions=1):
        """Initialize the Evaluate class."""
        if not isinstance(setting, (CEBulk, CECrystal)):
            msg = "setting must be CEBulk or CECrystal object"
            raise TypeError(msg)

        self.setting = setting
        if cf_names is None:
            self.cf_names = sorted(self.setting.all_cf_names)
        else:
            self.cf_names = sorted(cf_names)
        self.num_elements = setting.num_elements
        self.scoring_scheme = scoring_scheme
        if max_cluster_size is None:
            self.max_cluster_size = self.setting.max_cluster_size
        else:
            self.max_cluster_size = max_cluster_size
        if max_cluster_dia is None:
            self.max_cluster_dia = self.setting.max_cluster_dia
        else:
            self.max_cluster_dia = self._get_max_cluster_dia(max_cluster_dia)

        self.scheme = None
        self.scheme_string = None
        self.nsplits = nsplits
        self.num_repetitions = num_repetitions
        # Define the selection conditions
        self.select_cond = []
        if select_cond is None:
            self.select_cond = self.default_select_cond
        else:
            if isinstance(select_cond, list):
                self.select_cond += select_cond
            else:
                self.select_cond.append(select_cond)

        # Remove the cluster names that correspond to clusters larger than the
        # specified size and diameter.
        self._filter_cluster_name()

        self.cf_matrix = self._make_cf_matrix()
        self.e_dft, self.names, self.concs = self._get_dft_energy_per_atom()

        self.effective_num_data_pts = len(self.e_dft)
        self.weight_matrix = np.eye(len(self.e_dft))
        self._update_convex_hull_weight(min_weight)

        self.multiplicity_factor = self.setting.multiplicity_factor
        self.eci = None
        self.alpha = None
        self.e_pred_loo = None
        self.parallel = parallel
        if parallel:
            if num_core == "all":
                self.num_core = int(mp.cpu_count() / 2)
            else:
                self.num_core = int(num_core)

        self.set_fitting_scheme(fitting_scheme, alpha)

    @property
    def default_select_cond(self):
        return [('converged', '=', True),
                ('struct_type', '=', 'initial')]

    def set_fitting_scheme(self, fitting_scheme="ridge", alpha=1E-9):
        from clease.regression import LinearRegression
        allowed_fitting_schemes = ["ridge", "tikhonov", "lasso", "l1", "l2", "ols"]
        if isinstance(fitting_scheme, LinearRegression):
            self.scheme = fitting_scheme
        elif isinstance(fitting_scheme, str):
            fitting_scheme = fitting_scheme.lower()
            self.scheme_string = fitting_scheme
            if fitting_scheme not in allowed_fitting_schemes:
                raise ValueError("Fitting scheme has to be one of "
                                 "{}".format(allowed_fitting_schemes))
            if fitting_scheme in ["ridge", "tikhonov", "l2"]:
                from clease.regression import Tikhonov
                self.scheme = Tikhonov(alpha=alpha)
            elif fitting_scheme in ["lasso", "l1"]:
                from clease.regression import Lasso
                self.scheme = Lasso(alpha=alpha)
            else:
                # Perform ordinary least squares
                self.scheme = LinearRegression()
        else:
            raise ValueError("Fitting scheme has to be one of {} "
                             "or an LinearRegression instance."
                             "".format(allowed_fitting_schemes))

        # If the fitting scheme is changed, the ECIs computed are no
        # longer consistent with the scheme
        # By setting it to None, a new calculation is performed
        # when the ECIs are requested
        self.eci = None

        N = len(self.e_dft)

        # If user has supplied any data weighting, pass the
        # weight matrix to the fitting scheme
        if np.any(np.abs(self.weight_matrix - np.eye(N) > 1E-6)):
            self.scheme.weight_matrix = self.weight_matrix

    def _get_max_cluster_dia(self, max_cluster_dia):
        """Make max_cluster_dia in a numpy array form."""
        if isinstance(max_cluster_dia, (list, np.ndarray)):
            if len(max_cluster_dia) == self.max_cluster_size + 1:
                for i in range(2):
                    max_cluster_dia[i] = 0.
                max_cluster_dia = np.array(max_cluster_dia, dtype=float)
            elif len(max_cluster_dia) == self.max_cluster_size - 1:
                max_cluster_dia = np.array(max_cluster_dia, dtype=float)
                max_cluster_dia = np.insert(max_cluster_dia, 0, [0., 0.])
            else:
                raise ValueError("Invalid length for max_cluster_dia.")
        # max_cluster_dia is int or float
        elif isinstance(max_cluster_dia, (int, float)):
            max_cluster_dia *= np.ones(self.max_cluster_size - 1,
                                       dtype=float)
            max_cluster_dia = np.insert(max_cluster_dia, 0, [0., 0.])

        return max_cluster_dia

    def _update_convex_hull_weight(self, min_weight):
        """Weight structure according to similarity with the
           most similar structure on the Convex Hull."""

        if abs(min_weight - 1.0) < 1E-4:
            return

        from clease import ConvexHull
        cnv_hull = ConvexHull(self.setting.db_name,
                              select_cond=self.select_cond)
        hull = cnv_hull.get_convex_hull()

        cosine_sim = []
        for conc, energy in zip(self.concs, self.e_dft):
            sim = cnv_hull.cosine_similarity_convex_hull(conc, energy, hull)
            cosine_sim.append(sim)
        cosine_sim = np.array(cosine_sim)

        # Shift tha maximum value to 0
        cosine_sim -= np.max(cosine_sim)
        min_sim = np.min(cosine_sim)

        decay = np.log(min_weight) / min_sim

        self.weight_matrix = np.diag(np.exp(decay * cosine_sim))
        self.effective_num_data_pts = np.sum(self.weight_matrix)

    def get_eci(self):
        """Determine and return ECIs for a given alpha.

        This method also saves the last value of alpha used (self.alpha) and
        the corresponding ECIs (self.eci) such that ECIs are not calculated
        repeated if alpha value is unchanged.
        """
        self.eci = self.scheme.fit(self.cf_matrix, self.e_dft)
        return self.eci

    def get_eci_dict(self):
        """
        Determine cluster names and their corresponding ECI value and return
        them in a dictionary format."""
        self.get_eci()

        # sanity check
        if len(self.cf_names) != len(self.eci):
            raise ValueError('lengths of cf_names and ECIs are not same')

        i_nonzeros = np.nonzero(self.eci)[0]
        pairs = []
        for i, cname in enumerate(self.cf_names):
            if i not in i_nonzeros:
                continue
            pairs.append((cname, self.eci[i]))

        return dict(pairs)

    def save_eci(self, fname='eci'):
        """
        Save a dictionary of cluster names and their corresponding ECI value
        in JSON file format.

        Parameters:

        fname: str
            file name does not contain an extension (`.json`)
        """
        import json
        full_fname = fname + '.json'
        with open(full_fname, 'w') as outfile:
            json.dump(self.get_eci_dict(), outfile, indent=2,
                      separators=(",", ": "))

    def plot_fit(self, interactive=True, savefig=False, fname=None,
                 show_hull=True):
        """Plot calculated (DFT) and predicted energies for a given alpha.

        Paramters:

        alpha: int or float
            regularization parameter.

        savefig: bool
            - True: Save the plot with a file name specified in 'fname'.
                    Only works when interactive=False.
                    This option does not display figure.
            - False: Display figure without saving.

        fname: str
            file name of the figure (only used when savefig = True)

        show_hull: bool
            whether or not to show convex hull.
        """
        import matplotlib.pyplot as plt
        from clease import ConvexHull
        from clease.interactive_plot import ShowStructureOnClick

        if self.eci is None:
            self.get_eci()
        e_pred = self.cf_matrix.dot(self.eci)

        e_range = max(np.append(self.e_dft, e_pred)) - \
                  min(np.append(self.e_dft, e_pred))
        rmin = min(np.append(self.e_dft, e_pred)) - 0.05 * e_range
        rmax = max(np.append(self.e_dft, e_pred)) + 0.05 * e_range

        prefix = ""
        if fname is not None:
            prefix = fname.rpartition(".")[0]

        cv = None
        cv_name = "LOOCV"
        if self.scoring_scheme == "loocv":
            cv = self.loocv() * 1000.0
        elif self.scoring_scheme == "loocv_fast":
            cv = self.loocv_fast() * 1000.0
        elif self.scoring_scheme == "k-fold":
            cv = self.k_fold_cv() * 1000.0
            cv_name = "{}-fold".format(self.nsplits)
        t = np.arange(rmin - 10, rmax + 10, 1)
        fig = plt.figure()
        ax = fig.add_subplot(111)
        if self.effective_num_data_pts != len(self.e_dft):
            ax.set_title('Fit using {} data points. Eff. num. data points '
                         '{:.1f}'.format(self.e_dft.shape[0],
                                         self.effective_num_data_pts))
        else:
            ax.set_title('Fit using {} data points.'
                         ''.format(self.e_dft.shape[0]))

        if self.effective_num_data_pts != len(self.e_dft):
            w = np.diag(self.weight_matrix)
            im = ax.scatter(e_pred, self.e_dft, c=w)
            cb = fig.colorbar(im)
            cb.set_label("Weight")

            # Plot again with zero marker width to make the interactive
            # plot work
            ax.plot(e_pred, self.e_dft, 'o', mfc='none', color="black",
                    markeredgewidth=0.0)
        else:
            ax.plot(e_pred, self.e_dft, 'bo', mfc='none')
        ax.plot(t, t, 'r')
        ax.axis([rmin, rmax, rmin, rmax])
        ax.set_ylabel(r'$E_{DFT}$ (eV/atom)')
        ax.set_xlabel(r'$E_{pred}$ (eV/atom)')
        ax.text(0.95, 0.01,
                cv_name + " = {0:.3f} meV/atom \n"
                "RMSE = {1:.3f} meV/atom"
                "".format(cv, self.rmse() * 1000.0),
                verticalalignment='bottom', horizontalalignment='right',
                transform=ax.transAxes, fontsize=12)
        if self.e_pred_loo is not None:
            ax.plot(self.e_pred_loo, self.e_dft, 'ro', mfc='none')

        if interactive:
            lines = ax.get_lines()
            if self.e_pred_loo is None:
                data_points = [lines[0]]
            else:
                data_points = [lines[0], lines[2]]
            annotations = [self.names, self.names]
            db_name = self.setting.db_name
            ShowStructureOnClick(fig, ax, data_points, annotations, db_name)
        else:
            if savefig:
                plt.savefig(fname=fname)
            else:
                plt.show()

        # Create a plot with the residuals

        gridspec_kw = {"wspace": 0.0,
                       "width_ratios": [5, 1]}

        fig_residual, ax_res = plt.subplots(ncols=2, sharey=True,
                                            gridspec_kw=gridspec_kw)
        ax_residual = ax_res[0]
        ax_residual.set_title("LOO residual (o). Residual (v)")
        if self.e_pred_loo is None:
            loo_delta = None
        else:
            loo_delta = (self.e_dft - self.e_pred_loo) * 1000.0
        delta_e = (self.e_dft - e_pred) * 1000.0
        if self.effective_num_data_pts != len(self.e_dft) and \
                loo_delta is not None:
            im = ax_residual.scatter(self.e_dft, loo_delta, c=w)
            cb = fig_residual.colorbar(im)
            cb.set_label("Weight")

            # Plot again with zero with to make the interactive
            # plot work
            ax_residual.plot(self.e_dft, loo_delta, "o",
                             color="black", markeredgewidth=0.0, mfc="none")
        else:
            if loo_delta is not None:
                ax_residual.plot(self.e_dft, loo_delta, "o")

        ax_residual.plot(self.e_dft, delta_e, "v", mfc="none")

        ax_residual.axhline(0, ls="--")
        ax_residual.set_ylabel(r"$E_{DFT} - E_{pred}$ (meV/atom)")

        hist, bin_edges = np.histogram(delta_e, bins=30)
        h = bin_edges[1] - bin_edges[0]
        ax_res[1].barh(bin_edges[:-1], hist, height=h, color="#bdbdbd")

        ax_res[1].set_xlabel("# occ")
        ax_res[1].spines["right"].set_visible(False)
        ax_res[1].spines["top"].set_visible(False)

        if interactive:
            lines = ax_residual.get_lines()
            if loo_delta is not None:
                data_points = [lines[0], lines[1]]
                annotations = [self.names, self.names]
            else:
                data_points = [lines[0]]
                annotations = [self.names]
            ShowStructureOnClick(fig_residual, ax_residual, data_points,
                                 annotations, db_name)
        else:
            if savefig:
                fig_residual.savefig(prefix + "_residuals.png")
            else:
                plt.show()

        # Optionally show the convex hull
        if show_hull:
            cnv_hull = ConvexHull(self.setting.db_name,
                                  select_cond=self.select_cond)
            fig = cnv_hull.plot()

            concs = {k: [] for k in cnv_hull._unique_elem}
            for c in self.concs:
                for k in concs.keys():
                    concs[k].append(c.get(k, 0.0))
            form_en = [cnv_hull.get_formation_energy(c, e)
                       for c, e in zip(self.concs, e_pred.tolist())]
            cnv_hull.plot(fig=fig, concs=concs, energies=form_en, marker="x")
            fig.suptitle("Convex hull DFT (o), CE (x)")

            if savefig:
                fig.savefig(prefix + "_cnv_hull.png")
            else:
                plt.show()

    def alpha_CV(self, alpha_min=1E-7, alpha_max=1.0, num_alpha=10,
                 scale='log', logfile=None, fitting_schemes=None):
        """Calculate CV for a given range of alpha.

        In addition to calculating CV with respect to alpha, a logfile can be
        used to extend the range of alpha or to add more alpha values in a
        given range.

        Returns a list of alpha values, and a list of CV scores.

        Parameters:

        alpha_min: int or float
            minimum value of regularization parameter alpha.

        alpha_max: int or float
            maximum value of regularization parameter alpha.

        num_alpha: int
            number of alpha values to be used in the plot.

        scale: str
            - 'log'(default): alpha values are evenly spaced on a log scale.
            - 'linear': alpha values are evenly spaced on a linear scale.

        logfile: file object, str or None.
            - None: logging is disabled
            - str: a file with that name will be opened. If '-', stdout used.
            - file object: use the file object for logging

        fitting_schemes: None or array of instance of LinearRegression.

        Note: If the file with the same name exists, it first checks if the
              alpha value already exists in the logfile and evalutes the CV of
              the alpha values that are absent. The newly evaluated CVs are
              appended to the existing file.
        """
        from clease.regression import LinearRegression

        if fitting_schemes is None:
            if self.scheme_string is None:
                raise ValueError("No fitting scheme supplied!")
            if self.scheme_string in ["lasso", "l1"]:
                from clease.regression import Lasso
                fitting_schemes = Lasso.get_instance_array(
                    alpha_min, alpha_max, num_alpha=num_alpha, scale=scale)
            elif self.scheme_string in ["ridge", "l2", "tikhonov"]:
                from clease.regression import Tikhonov
                fitting_schemes = Tikhonov.get_instance_array(
                    alpha_min, alpha_max, num_alpha=num_alpha, scale=scale)

        for scheme in fitting_schemes:
            if not isinstance(scheme, LinearRegression):
                raise TypeError("Each entry in fitting_schemes should be an "
                                "instance of LinearRegression")
            elif not scheme.is_scalar():
                raise TypeError("plot_CV only supports the fitting schemes "
                                "with a scalar paramater.")

        # if the file exists, read the alpha values that are already evaluated.
        self._initialize_logfile(logfile)
        fitting_schemes = self._remove_existing_alphas(logfile,
                                                       fitting_schemes)

        for scheme in fitting_schemes:
            if not isinstance(scheme, LinearRegression):
                raise TypeError("Each entry in fitting_schemes should be an "
                                "instance of LinearRegression")
            elif not scheme.is_scalar():
                raise TypeError("plot_CV only supports the fitting schemes "
                                "with a scalar paramater.")

        # if the file exists, read the alpha values that are already evaluated.
        self._initialize_logfile(logfile)
        fitting_schemes = self._remove_existing_alphas(logfile,
                                                       fitting_schemes)

        # get CV scores
        alphas = []
        if self.parallel:
            workers = mp.Pool(self.num_core)
            args = [(self, scheme) for scheme in fitting_schemes]
            alphas = [s.get_scalar_parameter() for s in fitting_schemes]
            cv = workers.map(loocv_mp, args)
            cv = np.array(cv)
            workers.close()
        else:
            cv = np.ones(len(fitting_schemes))
            for i, scheme in enumerate(fitting_schemes):
                self.set_fitting_scheme(fitting_scheme=scheme)
                if self.scoring_scheme == "loocv":
                    cv[i] = self.loocv()
                elif self.scoring_scheme == "loocv_fast":
                    cv[i] = self.loocv_fast()
                elif self.scoring_scheme == "k-fold":
                    cv[i] = self.k_fold_cv()
                num_eci = len(np.nonzero(self.get_eci())[0])
                alpha = scheme.get_scalar_parameter()
                alphas.append(alpha)
                logger.info('{:.10f}\t {}\t {:.10f}'.format(alpha, num_eci,
                                                            cv[i]))

        return alphas, cv

    def plot_CV(self, alpha_min=1E-7, alpha_max=1.0, num_alpha=10, scale='log',
                logfile=None, fitting_schemes=None, savefig=False, fname=None):
        """Plot CV for a given range of alpha.

        In addition to plotting CV with respect to alpha, logfile can be used
        to extend the range of alpha or add more alpha values in a given range.
        Returns an alpha value that leads to the minimum CV score within the
        pool of evaluated alpha values.

        Parameters:

        alpha_min: int or float
            minimum value of regularization parameter alpha.

        alpha_max: int or float
            maximum value of regularization parameter alpha.

        num_alpha: int
            number of alpha values to be used in the plot.

        scale: str
            - 'log'(default): alpha values are evenly spaced on a log scale.
            - 'linear': alpha values are evenly spaced on a linear scale.

        logfile: file object, str or None
            - None: logging is disabled
            - str: a file with that name will be opened. If '-', stdout used.
            - file object: use the file object for logging

        fitting_schemes: None or array of instance of LinearRegression

        savefig: bool
            - True: Save the plot with a file name specified in 'fname'. This
                    option does not display figure.
            - False: Display figure without saving.

        fname: str
            file name of the figure (only used when savefig = True)

        Note: If the file with the same name exists, it first checks if the
              alpha value already exists in the logfile and evalutes the CV of
              the alpha values that are absent. The newly evaluated CVs are
              appended to the existing file.
        """
        import matplotlib.pyplot as plt

        alphas, cv = self.alpha_CV(alpha_min=alpha_min, alpha_max=alpha_max,
                                   num_alpha=num_alpha, scale=scale,
                                   logfile=logfile,
                                   fitting_schemes=fitting_schemes)
        # --------------- #
        # Generate a plot #
        # --------------- #
        # if logfile is present, read all entries from the file
        if logfile is not None and logfile != '-':
            alphas, cv = self._get_alphas_cv_from_file(logfile)

        # get the minimum CV score and the corresponding alpha value
        ind = cv.argmin()
        min_alpha = alphas[ind]
        min_cv = cv[ind]

        fig = plt.figure()
        ax = fig.add_subplot(111)
        ax.set_title('CV score vs. alpha')
        ax.semilogx(alphas, cv * 1000)
        ax.semilogx(min_alpha, min_cv * 1000, 'bo', mfc='none')
        ax.set_ylabel('CV score (meV/atom)')
        ax.set_xlabel('alpha')
        ax.text(0.65, 0.01, "min. CV score:\n"
                "alpha = {0:.10f} \n"
                "CV = {1:.3f} meV/atom".format(min_alpha, min_cv * 1000.0),
                verticalalignment='bottom', horizontalalignment='left',
                transform=ax.transAxes, fontsize=10)
        if savefig:
            plt.savefig(fname=fname)
        else:
            plt.show()
        return min_alpha

    def _get_alphas_cv_from_file(self, logfile):
        alphas = []
        cv = []
        with open(logfile) as log:
            next(log)
            for line in log:
                alphas.append(float(line.split()[0]))
                cv.append(float(line.split()[-1]))
            alphas = np.array(alphas)
            cv = np.array(cv)
            # sort alphas and cv based on the values of alphas
            ind = alphas.argsort()
            alphas = alphas[ind]
            cv = cv[ind]
        return alphas, cv

    def _remove_existing_alphas(self, logfile, fitting_schemes):
        if not isinstance(logfile, str):
            return fitting_schemes
        elif logfile == "-":
            return fitting_schemes

        existing_alpha = []
        with open(logfile) as f:
            lines = f.readlines()
        for line_num, line in enumerate(lines):
            if line_num == 0:
                continue
            existing_alpha.append(float(line.split()[0]))
        schemes = []
        for scheme in fitting_schemes:
            exists = np.isclose(existing_alpha, scheme.get_scalar_parameter(),
                                atol=1E-9).any()
            if not exists:
                schemes.append(scheme)
        return schemes

    def _initialize_logfile(self, logfile):
        # logfile setup
        if isinstance(logfile, basestring):
            if logfile == '-':
                handler = lg.StreamHandler(sys.stdout)
                handler.setLevel(lg.INFO)
                logger.addHandler(handler)
            else:
                handler = MultiprocessHandler(logfile)
                handler.setLevel(lg.INFO)
                logger.addHandler(handler)
                # create a log file and make a header line if the file does not
                # exist.
                if os.stat(logfile).st_size == 0:
                    logger.info("alpha \t\t # ECI \t CV")

    def plot_ECI(self, ignore_sizes=[0], interactive=True):
        """Plot the all the ECI.

        Parameters:

        ignore_sizes: list of ints
            Sizes listed in this list will not be plotted.
            Default is to ignore the emptry cluster.

        interactive: bool
            If ``True``, one can interact with the plot using mouse.
        """
        import matplotlib.pyplot as plt
        from clease.interactive_plot import InteractivePlot

        if self.eci is None:
            self.get_eci()
        distances = self._distance_from_names()

        # Structure the ECIs in terms by size
        eci_by_size = {}
        for name, d, eci in zip(self.cf_names, distances, self.eci):
            size = int(name[1])
            if size not in eci_by_size.keys():
                eci_by_size[size] = {"d": [], "eci": [], "name": []}
            eci_by_size[size]["d"].append(d)
            eci_by_size[size]["eci"].append(eci)
            eci_by_size[size]["name"].append(name)

        fig = plt.figure()
        ax = fig.add_subplot(1, 1, 1)
        ax.axhline(0.0, ls="--", color="grey")
        markers = ["o", "v", "x", "D", "^", "h", "s", "p"]
        annotations = []
        lines = []
        for size, data in eci_by_size.items():
            if size in ignore_sizes:
                continue
            data["d"] = np.array(data["d"])
            data["eci"] = np.array(data["eci"])
            sort_index = np.argsort(data["d"])
            data["d"] = data["d"][sort_index]
            data["eci"] = data["eci"][sort_index]
            annotations.append([data["name"][indx] for indx in sort_index])
            mrk = markers[size % len(markers)]
            line = ax.plot(data["d"], data["eci"],
                           label="{}-body".format(size), marker=mrk,
                           mfc="none", ls="", markersize=8)
            lines.append(line[0])

        ax.xaxis.set_major_locator(plt.MultipleLocator(1))
        ax.set_xlabel("Cluster diameter ($n^{th}$ nearest neighbor)")
        ax.set_ylabel("ECI (eV/atom)")
        ax.legend()
        if interactive:
            # Note: Internally this calls plt.show()
            InteractivePlot(fig, ax, lines, annotations)
        else:
            plt.show()

    def _distance_from_names(self):
        """Get a list with all the distances for each name."""
        dists = []
        for name in self.cf_names:
            if name == "c0" or name.startswith("c1"):
                dists.append(0)
                continue
            dist_str = name.split("_")[1][1:]
            dists.append(int(dist_str))
        return dists

    def mae(self):
        """Calculate mean absolute error (MAE) of the fit."""
        if self.eci is None:
            self.get_eci()
        e_pred = self.cf_matrix.dot(self.eci)
        delta_e = self.e_dft - e_pred
        w = np.diag(self.weight_matrix)
        delta_e *= w
        return sum(np.absolute(delta_e)) / self.effective_num_data_pts

    def rmse(self):
        """Calculate root-mean-square error (RMSE) of the fit."""
        if self.eci is None:
            self.get_eci()
        e_pred = self.cf_matrix.dot(self.eci)
        delta_e = self.e_dft - e_pred

        w = np.diag(self.weight_matrix)
        rmse_sq = np.sum(w * delta_e**2)
        rmse_sq /= self.effective_num_data_pts
        return np.sqrt(rmse_sq)

    def loocv_fast(self):
        """CV score based on the method in J. Phase Equilib. 23, 348 (2002).

        This method has a computational complexity of order n^1.
        """
        # For each structure i, predict energy based on the ECIs determined
        # using (N-1) structures and the parameters corresponding to the
        # structure i.
        # CV^2 = N^{-1} * Sum((E_DFT-E_pred) / (1 - X_i (X^T X)^{-1} X_u^T))^2
        if not self.scheme.support_fast_loocv:
            return self.loocv()

        if self.eci is None:
            self.get_eci()
        e_pred = self.cf_matrix.dot(self.eci)
        delta_e = self.e_dft - e_pred
        cfm = self.cf_matrix
        # precision matrix
        prec = self.scheme.precision_matrix(cfm)
        delta_e_loo = delta_e / (1 - np.diag(cfm.dot(prec).dot(cfm.T)))
        self.e_pred_loo = self.e_dft - delta_e_loo
        w = np.diag(self.weight_matrix)
        cv_sq = np.sum(w * delta_e_loo**2)

        cv_sq /= self.effective_num_data_pts
        return np.sqrt(cv_sq)

    def loocv(self):
        """Determine the CV score for the Leave-One-Out case."""
        cv_sq = 0.
        e_pred_loo = []
        for i in range(self.cf_matrix.shape[0]):
            eci = self._get_eci_loo(i)
            e_pred = self.cf_matrix[i][:].dot(eci)
            delta_e = self.e_dft[i] - e_pred
            cv_sq += self.weight_matrix[i, i] * (delta_e)**2
            e_pred_loo.append(e_pred)
        # cv_sq /= self.cf_matrix.shape[0]
        cv_sq /= self.effective_num_data_pts
        self.e_pred_loo = e_pred_loo
        return np.sqrt(cv_sq)

    def k_fold_cv(self):
        """Determine the k-fold cross validation."""
        from clease.tools import split_dataset
        avg_score = 0.0
        for _ in range(self.num_repetitions):
            partitions = split_dataset(self.cf_matrix, self.e_dft,
                                       nsplits=self.nsplits)
            scores = []
            for part in partitions:
                eci = self.scheme.fit(part["train_X"], part["train_y"])
                e_pred = part["validate_X"].dot(eci)
                scores.append(np.mean((e_pred - part["validate_y"])**2))
            avg_score += np.sqrt(np.mean(scores))
        return avg_score / self.num_repetitions

    def _get_eci_loo(self, i):
        """Determine ECI values for the Leave-One-Out case.

        Eliminate the ith row of the cf_matrix when determining the ECIs.
        Returns the determined ECIs.

        Parameters:

        i: int
            iterator passed from the self.loocv method.
        """
        cfm = np.delete(self.cf_matrix, i, 0)
        e_dft = np.delete(self.e_dft, i, 0)
        eci = self.scheme.fit(cfm, e_dft)
        return eci

    def _filter_cluster_name(self):
        """Filter the cluster names based on size and diameter."""
        if self.max_cluster_size is None and self.max_cluster_dia is None:
            return

        filtered_cnames = []
        for name in self.cf_names:
            size = int(name[1])
            if size < 2:
                dia = -1
            else:
                prefix = name.rpartition("_")[0]
                cluster = self.setting.cluster_list.get_by_name(prefix)[0]
                dia = cluster.diameter
            if (size <= self.max_cluster_size
                    and dia < self.max_cluster_dia[size]):
                filtered_cnames.append(name)
        self.cf_names = filtered_cnames

    def _make_cf_matrix(self):
        """Return a matrix containing the correlation functions.

        Only selects all of the converged structures by default, but further
        constraints can be imposed using *select_cond* argument in the
        initialization step.
        """
        cfm = []
        db = connect(self.setting.db_name)
        tab_name = "{}_cf".format(self.setting.bf_scheme.name)
        for row in db.select(self.select_cond):
            cfm.append([row[tab_name][x] for x in self.cf_names])
        return np.array(cfm, dtype=float)

    def _get_dft_energy_per_atom(self):
        """Retrieve DFT energy and convert it to eV/atom unit."""
        e_dft = []
        names = []
        concentrations = []
        db = connect(self.setting.db_name)
        for row in db.select(self.select_cond):
            final_struct_id = row.get("final_struct_id", -1)
            if final_struct_id >= 0:
                # New format where energy is stored in a separate DB entry
                energy = db.get(id=final_struct_id).energy
            else:
                # Old format where the energy is stored in the init structure
                energy = row.energy
            e_dft.append(energy / row.natoms)
            names.append(row.name)

            count = row.count_atoms()
            for k in count.keys():
                count[k] /= row.natoms
            concentrations.append(count)
        return np.array(e_dft), names, concentrations

    def _get_cf_from_atoms_row(self, row):
        """Obtain the correlation functions from an Atoms Row object."""
        return [row[x] for x in self.cf_names]

    def generalization_error(self, validation_id):
        """
        Estimate the generalization error to new datapoints

        Parameters:

        validation_ids: list of int
            List with IDs to leave out of the dataset
        """

        db = connect(self.setting.db_name)

        mse = 0.0
        count = 0
        for uid in validation_id:
            row = db.get(id=uid)
            cf = self._get_cf_from_atoms_row(row)

            pred = self.eci.dot(cf)

            if row.get('final_struct_id', -1) != -1:
                energy = db.get(id=row.get('final_struct_id'))
            else:
                energy = row.energy
            mse += (energy / row.natoms - pred)**2
            count += 1

        if count == 0:
            return 0.0
        return np.sqrt(mse / count) * 1000.0

    def export_dataset(self, fname):
        """
        Export the dataset used to fit a model y = Xc where y is typically the
        DFT energy per atom and c is the unknown ECIs. This function exports
        the data to a csv file with the following format

        # ECIname_1, ECIname_2, ..., ECIname_n, E_DFT
        0.1, 0.4, ..., -0.6, -2.0
        0.3, 0.2, ..., -0.9, -2.3

        thus each row in the file contains the correlation function values and
        the corresponding DFT energy value.

        Parameter:

        fname: str
            Filename to write to. Typically this should end with .csv
        """
        header = ','.join(self.cf_names) + ',E_DFT (eV/atom)'
        data = np.hstack((self.cf_matrix,
                          np.reshape(self.e_dft, (len(self.e_dft), -1))))
        np.savetxt(fname, data, delimiter=",", header=header)
        _logger("Dataset exported to {}".format(fname))

    def get_cv_score(self):
        """
        Calculate the CV score according to the selected scheme
        """
        if self.scoring_scheme == "loocv":
            cv = self.loocv() * 1000.0
        elif self.scoring_scheme == "loocv_fast":
            cv = self.loocv_fast() * 1000.0
        elif self.scoring_scheme == "k-fold":
            cv = self.k_fold_cv() * 1000.0
        return cv


def loocv_mp(args):
    """Need to wrap this function in order to use it with multiprocessing.

    Parameters:

    args: Tuple where the first entry is an instance of Evaluate
        and the second is the penalization value
    """
    evaluator = args[0]
    scheme = args[1]
    evaluator.set_fitting_scheme(fitting_scheme=scheme)
    alpha = scheme.get_scalar_parameter()

    if evaluator.scoring_scheme == "loocv":
        cv = evaluator.loocv()
    elif evaluator.scoring_scheme == "loocv_fast":
        cv = evaluator.loocv_fast()
    elif evaluator.scoring_scheme == "k-fold":
        cv = evaluator.k_fold_cv()
    num_eci = len(np.nonzero(evaluator.get_eci())[0])
    logger.info('{:.10f}\t {}\t {:.10f}'.format(alpha, num_eci, cv))
    return cv
